-- $Id: stores.lua 82 2012-09-20 04:30:24Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

-- Store entities generated from T2's st_info.txt.

local function weighted_store_filter(t)
  for _ = 1, 50 do
    local f = t[rng.range(1, #t)]
    if not f.prob or rng.percent(f.prob) then return f end
  end
  -- If we didn't get one after fifty tries, force the issue and randomly
  -- pick one of the 100%-probability ones.
  local t100 = {}
  for _, f in ipairs(t) do
    if not f.prob or f.prob >= 100 then table.insert(t100, f) end
  end
  if #t100 == 0 then return nil end
  return t100[rng.range(1, #t100)]
end

local function black_market_filter()
  -- No artifact bases or specially generated objects; otherwise, anything
  -- goes.
  return {
    not_properties = { 'special_gene', 'insta_art', 'norm_art' },
    id = true,
  }
end

local stock_base = {
  GENERAL_STORE = {
    misc = { corpse=true, oil=true, spike=true, bottle=true, },
    food = true,
    light = true,
    ammo = true,
    tool = 'digger',
    armor = 'cloak',
  },
  ARMOURY = {
    armor = true,
  },
  WEAPONSMITH = {
    ammo = true,
    weapon = true,
    shooter = true,
  },
  TEMPLE = {
    magic = {
      spellbook = function(o)
	-- TODO Allow priest spell books.
	return false
      end
    },
    utility = { scroll=true, potion=true, },
    weapon = function(o)
      -- TODO Allow blessed weapons.
      return o.subtype == 'hafted' or o.subtype == 'magestaff'
    end
  },
  ALCHEMY_SHOP = {
    utility = { scroll=true, potion=true, alchemic=true, bottle=true, },
  },
  MAGIC_SHOP = {
    magic = {
      spellbook = function(o)
	-- TODO Allow mage spell books.
	return false
      end
    },
    jewelry = { ring=true, amulet=true, },
    utility = {
      staff=true, wand=true, rod=true, rodtip=true, scroll=true, potion=true,
    },
    weapon = 'magestaff',
    misc = 'junkart',
  },
  BLACK_MARKET = function(o) return true end,
  BOOK_STORE = {
    magic = 'spellbook',
    weapon = function(o) return o.is_demongear end,
    armor = function(o) return o.is_demongear end,
  },
  PET_SHOP = {
    misc = 'egg',
  },
}

local function allow_sell(tag, o)
  local base = stock_base[tag]
  if not base then return false end
  if type(base) == 'table' and base[o.type] then
    local x = base[o.type]
    if type(x) == 'boolean' then return true end
    if type(x) == 'string' then return x == o.subtype end
    if type(x) == 'function' then return x(o) end
    if type(x) == 'table' and x[o.subtype] then
      local xx = base[o.type][o.subtype]
      if type(xx) == 'boolean' then return true end
      if type(xx) == 'function' then return xx(o) end
    end
  end
  if type(base) == 'function' then return base(o) end
  return false
end

-- Annoyingly, the store.filters() function can't be a :method() with an
-- implicit self argument, so it can't refer to any fields in the store
-- entity.  Instead we declare the filter tables as local variables for
-- store.filters() to pass to weighted_store_filter() above.
local filters_GENERAL_STORE = {
  { defined='WOODEN_TORCH', id='true' },
  { defined='BRASS_LANTERN', id='true', prob=95 },
  { defined='FLASK_OF_OIL', id='true', prob=95 },
  { defined='RATION_OF_FOOD', id='true' },
  { defined='HARD_BISCUIT', id='true' },
  { defined='STRIP_OF_VENISON', id='true', prob=90 },
  { defined='PINT_OF_FINE_WINE', id='true', prob=70 },
  { defined='PINT_OF_FINE_ALE', id='true', prob=80 },
  { defined='SHOVEL', id='true', prob=60 },
  { defined='PICK', id='true', prob=50 },
  { defined='IRON_SPIKE', id='true' },
  { defined='IRON_SHOT', id='true', prob=70 },
  { defined='BOLT', id='true', prob=70 },
  { defined='ARROW', id='true', prob=70 },
  { defined='CLOAK', id='true', prob=98 },
  { defined='FUR_CLOAK', id='true', prob=46 },
}
newEntity {
  define_as = 'GENERAL_STORE',
  name = "General Store",
  display = '1',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'BILBO', 'RAISTLIN', 'SULTAN', 'LYAR_EL' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  store = {
    filters = function() return weighted_store_filter(filters_GENERAL_STORE) end,
    max_stock = 24,
    actor_filter = function(o) return allow_sell('GENERAL_STORE', o) end,
  },
}

local filters_ARMOURY = {
  { defined='PAIR_OF_SOFT_LEATHER_BOOTS', id='true', prob=70 },
  { defined='PAIR_OF_HARD_LEATHER_BOOTS', id='true', prob=60 },
  { defined='HARD_LEATHER_CAP', id='true', prob=80 },
  { defined='METAL_CAP', id='true', prob=70 },
  { defined='IRON_HELM', id='true', prob=65 },
  { defined='ROBE', id='true' },
  { defined='SOFT_LEATHER_ARMOUR', id='true', prob=90 },
  { defined='SOFT_STUDDED_LEATHER', id='true', prob=90 },
  { defined='HARD_LEATHER_ARMOUR', id='true', prob=90 },
  { defined='HARD_STUDDED_LEATHER', id='true', prob=85 },
  { defined='LEATHER_SCALE_MAIL', id='true', prob=90 },
  { defined='METAL_SCALE_MAIL', id='true', prob=80 },
  { defined='CHAIN_MAIL', id='true', prob=75 },
  { defined='AUGMENTED_CHAIN_MAIL', id='true', prob=80 },
  { defined='METAL_BRIGANDINE_ARMOUR', id='true', prob=75 },
  { defined='BAR_CHAIN_MAIL', id='true', prob=68 },
  { defined='SET_OF_LEATHER_GLOVES', id='true' },
  { defined='SET_OF_GAUNTLETS', id='true', prob=80 },
  { defined='SMALL_LEATHER_SHIELD', id='true' },
  { defined='LARGE_LEATHER_SHIELD', id='true', prob=90 },
  { defined='SMALL_METAL_SHIELD', id='true', prob=80 },
}
newEntity {
  define_as = 'ARMOURY',
  name = "Armoury",
  display = '2',
  color = colors.C_s,
  owner = resolvers.store_owner{ 'KON_DAR', 'DARG_LOW', 'DECADO', 'WIELAND' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  store = {
    filters = function() return weighted_store_filter(filters_ARMOURY) end,
    max_stock = 24,
    actor_filter = function(o) return allow_sell('ARMOURY', o) end,
  },
}

local filters_WEAPONSMITH = {
  { defined='DAGGER', id='true' },
  { defined='MAIN_GAUCHE', id='true' },
  { defined='RAPIER', id='true' },
  { defined='SMALL_SWORD', id='true' },
  { defined='SHORT_SWORD', id='true' },
  { defined='SABRE', id='true' },
  { defined='CUTLASS', id='true' },
  { defined='TULWAR', id='true' },
  { defined='BROAD_SWORD', id='true' },
  { defined='LONG_SWORD', id='true' },
  { defined='SCIMITAR', id='true' },
  { defined='KATANA', id='true' },
  { defined='BASTARD_SWORD', id='true' },
  { defined='SPEAR', id='true' },
  { defined='AWL_PIKE', id='true' },
  { defined='TRIDENT', id='true' },
  { defined='PIKE', id='true' },
  { defined='BEAKED_AXE', id='true' },
  { defined='BROAD_AXE', id='true' },
  { defined='LANCE', id='true' },
  { defined='BATTLE_AXE', id='true' },
  { defined='HATCHET', id='true' },
  { defined='SLING', id='true' },
  { defined='SHORT_BOW', id='true' },
  { defined='LONG_BOW', id='true' },
  { defined='LIGHT_CROSSBOW', id='true' },
  { defined='IRON_SHOT', id='true' },
  { defined='ARROW', id='true' },
  { defined='BOLT', id='true' },
  { defined='WHIP', id='true' },
  { defined='SMALL_WOODEN_BOOMERANG', id='true' },
}
newEntity {
  define_as = 'WEAPONSMITH',
  name = "Weaponsmith",
  display = '3',
  color = colors.C_w,
  owner = resolvers.store_owner{ 'ARNOLD', 'ARNDAL', 'EDDIE', 'OGLIGN' },
  -- actions:
  --   Compare weapons (23)
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  store = {
    filters = function() return weighted_store_filter(filters_WEAPONSMITH) end,
    max_stock = 24,
    actor_filter = function(o) return allow_sell('WEAPONSMITH', o) end,
  },
}

local filters_TEMPLE = {
  { defined='QUARTERSTAFF', id='true' },
  { defined='MACE', id='true' },
  { defined='WAR_HAMMER', id='true' },
  { defined='LUCERNE_HAMMER', id='true' },
  { defined='MORNING_STAR', id='true' },
  { defined='FLAIL', id='true' },
  { defined='LEAD_FILLED_MACE', id='true' },
  { defined='SCRL_REMOVE_CURSE', id='true' },
  { defined='SCRL_BLESSING', id='true' },
  { defined='SCRL_HOLY_CHANT', id='true' },
  { defined='POT_HEROISM', id='true' },
  { defined='SCRL_WORD_OF_RECALL', id='true' },
  { defined='SCRL_WORD_OF_RECALL', id='true' },
  { defined='SCRL_WORD_OF_RECALL', id='true' },
  { defined='POT_CURE_LIGHT_WOUNDS', id='true' },
  { defined='POT_CURE_SERIOUS_WOUNDS', id='true' },
  { defined='POT_CURE_SERIOUS_WOUNDS', id='true' },
  { defined='POT_CURE_CRITICAL_WOUNDS', id='true' },
  { defined='POT_CURE_CRITICAL_WOUNDS', id='true' },
  { defined='POT_RESTORE_LIFE_LEVELS', id='true' },
  { defined='POT_RESTORE_LIFE_LEVELS', id='true' },
  { defined='POT_RESTORE_LIFE_LEVELS', id='true' },
  { defined='WHIP', id='true' },
  { defined='MACE', id='true' },
  { defined='BALL_AND_CHAIN', id='true' },
  { defined='WAR_HAMMER', id='true' },
  { defined='SCRL_WORD_OF_RECALL', id='true' },
  { defined='SCRL_WORD_OF_RECALL', id='true' },
  { defined='SCRL_WORD_OF_RECALL', id='true' },
  { defined='POT_CURE_CRITICAL_WOUNDS', id='true' },
  { defined='POT_CURE_CRITICAL_WOUNDS', id='true' },
  { defined='POT_RESTORE_LIFE_LEVELS', id='true' },
  { defined='POT_RESTORE_LIFE_LEVELS', id='true' },
  { defined='POT_RESTORE_LIFE_LEVELS', id='true' },
  { defined='SCRL_REMOVE_CURSE', id='true' },
  { defined='SCRL_REMOVE_CURSE', id='true' },
  { defined='SCRL_XTRA_REMOVE_CURSE', id='true' },
  { defined='SCRL_XTRA_REMOVE_CURSE', id='true' },
}
newEntity {
  define_as = 'TEMPLE',
  name = "Temple",
  display = '4',
  color = colors.C_g,
  owner = resolvers.store_owner{ 'LUDWIG', 'GUNNAR', 'TORIN', 'SARASTRO' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  store = {
    filters = function() return weighted_store_filter(filters_TEMPLE) end,
    max_stock = 24,
    actor_filter = function(o) return allow_sell('TEMPLE', o) end,
  },
}

local filters_ALCHEMY_SHOP = {
  { defined='SCRL_ENCHANT_WEAPON_TO_HIT', id='true' },
  { defined='SCRL_ENCHANT_WEAPON_TO_DAM', id='true' },
  { defined='SCRL_ENCHANT_ARMOUR', id='true' },
  { defined='STAFF_IDENTIFY', id='true' },
  { defined='STAFF_IDENTIFY', id='true' },
  { defined='STAFF_IDENTIFY', id='true' },
  { defined='STAFF_IDENTIFY', id='true' },
  { defined='ESS_LIGHT', id='true' },
  { defined='SCRL_PHASE_DOOR', id='true' },
  { defined='SCRL_PHASE_DOOR', id='true' },
  { defined='SCRL_TELEPORTATION', id='true' },
  { defined='SCRL_MONSTER_CONFUSION', id='true' },
  { defined='SCRL_MAGIC_MAPPING', id='true' },
  { defined='SCRL_TREASURE_DETECTION', id='true' },
  { defined='SCRL_OBJECT_DETECTION', id='true' },
  { defined='SCRL_TRAP_DETECTION', id='true' },
  { defined='POT_DETECT_INVISIBLE', id='true' },
  { defined='SCRL_RECHARGING', id='true' },
  { defined='SCRL_SATISFY_HUNGER', id='true' },
  { defined='SCRL_WORD_OF_RECALL', id='true' },
  { defined='SCRL_WORD_OF_RECALL', id='true' },
  { defined='SCRL_WORD_OF_RECALL', id='true' },
  { defined='SCRL_WORD_OF_RECALL', id='true' },
  { defined='SCRL_TELEPORTATION', id='true' },
  { defined='SCRL_TELEPORTATION', id='true' },
  { defined='POT_RESTORE_STRENGTH', id='true' },
  { defined='POT_RESTORE_INTELLIGENCE', id='true' },
  { defined='POT_RESTORE_WISDOM', id='true' },
  { defined='POT_RESTORE_DEXTERITY', id='true' },
  { defined='POT_RESTORE_CONSTITUTION', id='true' },
  { defined='POT_RESTORE_CHARISMA', id='true' },
  { defined='STAFF_IDENTIFY', id='true' },
  { defined='SCRL_XTRA_IDENTIFY', id='true' },
  { defined='SCRL_XTRA_IDENTIFY', id='true' },
  { defined='SCRL_XTRA_IDENTIFY', id='true' },
  { defined='SCRL_XTRA_IDENTIFY', id='true' },
  { defined='SCRL_XTRA_IDENTIFY', id='true' },
  { defined='SCRL_XTRA_IDENTIFY', id='true' },
  { defined='ESS_LIGHT', id='true' },
  { defined='POT_RESTORE_STRENGTH', id='true' },
  { defined='POT_RESTORE_INTELLIGENCE', id='true' },
  { defined='POT_RESTORE_WISDOM', id='true' },
  { defined='POT_RESTORE_DEXTERITY', id='true' },
  { defined='POT_RESTORE_CONSTITUTION', id='true' },
  { defined='POT_RESTORE_CHARISMA', id='true' },
  { defined='SCRL_ENCHANT_ARMOUR', id='true' },
  { defined='SCRL_ENCHANT_ARMOUR', id='true' },
  { defined='SCRL_RECHARGING', id='true' },
  { defined='SCRL_SATISFY_HUNGER', id='true' },
  { defined='SCRL_SATISFY_HUNGER', id='true' },
  { defined='SCRL_SATISFY_HUNGER', id='true' },
}
newEntity {
  define_as = 'ALCHEMY_SHOP',
  name = "Alchemy shop",
  display = '5',
  color = colors.C_b,
  owner = resolvers.store_owner{ 'MAUSER', 'WIZZLE', 'MIDAS', 'JA_FAR' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  store = {
    filters = function() return weighted_store_filter(filters_ALCHEMY_SHOP) end,
    max_stock = 24,
    actor_filter = function(o) return allow_sell('ALCHEMY_SHOP', o) end,
  },
}

local filters_MAGIC_SHOP = {
  { defined='RUNE_PROTECTION', id='true' },
  { defined='RING_LEVITATION', id='true' },
  { defined='RUNE_PROTECTION', id='true' },
  { defined='RING_FIRE_RESISTANCE', id='true' },
  { defined='RING_COLD_RESISTANCE', id='true' },
  { defined='POT_CHARISMA', id='true' },
  { defined='AMULET_SLOW_DIGESTION', id='true' },
  { defined='AMULET_SLOW_DIGESTION', id='true' },
  { defined='AMULET_ACID_RESISTANCE', id='true' },
  { defined='AMULET_LIGHTNING_RESISTANCE', id='true' },
  { defined='AMULET_SEARCHING', id='true' },
  { defined='POT_CURE_LIGHT_WOUNDS', id='true' },
  { defined='TIP_PROBING', id='true' },
  { defined='WOODEN_ROD_OF_', id='true', prob=25 },
  { defined='BOOK_OF_BEGINNER_CANTRIPS', id='true' },
  { defined='WAND_MANATHRUST', id='true' },
  { defined='WAND_NOXIOUS_CLOUD', id='true' },
  { defined='WAND_DIG', id='true' },
  { defined='WAND_MAGELOCK', id='true' },
  { defined='WAND_CONFUSE', id='true' },
  { defined='STAFF_GLOBE_OF_LIGHT', id='true' },
  { defined='STAFF_REMOVE_CURSES', id='true' },
  { defined='STAFF_DISARM', id='true' },
  { defined='STAFF_TELEPORTATION', id='true' },
  { defined='STAFF_IDENTIFY', id='true' },
  { defined='STAFF_SENSE_HIDDEN', id='true', prob=60 },
  { defined='STAFF_REVEAL_WAYS', id='true', prob=60 },
  { defined='STAFF_SENSE_MONSTERS', id='true', prob=60 },
}
newEntity {
  define_as = 'MAGIC_SHOP',
  name = "Magic shop",
  display = '6',
  color = colors.C_r,
  owner = resolvers.store_owner{ 'ARIEL', 'BUGGERBY', 'INGLORIAN', 'LUTHIEN' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  store = {
    filters = function() return weighted_store_filter(filters_MAGIC_SHOP) end,
    max_stock = 24,
    actor_filter = function(o) return allow_sell('MAGIC_SHOP', o) end,
  },
}

newEntity {
  define_as = 'BLACK_MARKET',
  name = "Black Market",
  display = '7',
  color = colors.C_D,
  owner = resolvers.store_owner{ 'GARY', 'HISTOR', 'ZORAK', 'TIPO' },
  -- actions:
  --   Get share of stolen gold (30)
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL ALL_ITEM ]
  store = {
    filters = black_market_filter,
    max_stock = 24,
    actor_filter = function(o) return allow_sell('BLACK_MARKET', o) end,
  },
}

newEntity {
  define_as = 'HOME',
  name = "Home",
  display = '8',
  color = colors.C_y,
  owner = resolvers.store_owner{ 'BILBO' },
  -- actions:
  --   Drop an item (54)
  --   Get an item (55)
  --   Examine an item (3)
  store = {
    max_stock = 24,
  },
}

local filters_BOOK_STORE = {
  { defined='BOOK_OF_BEGINNER_CANTRIPS', id='true' },
  { defined='BOOK_OF_BEGINNER_CANTRIPS', id='true' },
  { defined='SPELLBOOK_OF__', id='true' },
  { defined='SPELLBOOK_OF__', id='true' },
  { defined='SPELLBOOK_OF__', id='true' },
  { defined='SPELLBOOK_OF__', id='true' },
  { defined='SPELLBOOK_OF__', id='true' },
  { defined='SPELLBOOK_OF__', id='true' },
}
newEntity {
  define_as = 'BOOK_STORE',
  name = "Book Store",
  display = '9',
  color = colors.C_o,
  owner = resolvers.store_owner{ 'DOLAF', 'ODNAR', 'GANDAR', 'RO_SHA' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  store = {
    filters = function() return weighted_store_filter(filters_BOOK_STORE) end,
    max_stock = 24,
    actor_filter = function(o) return allow_sell('BOOK_STORE', o) end,
  },
}

local filters_PET_SHOP = {
  { defined='EGG', id='true' },
  { defined='SCRL_SUMMON_NEVER_MOVING_PET', id='true' },
  { defined='HARD_BISCUIT', id='true' },
}
newEntity {
  define_as = 'PET_SHOP',
  name = "Pet Shop",
  display = '+',
  color = colors.C_b,
  owner = resolvers.store_owner{ 'N_RAK', 'ESPERION', 'FLARIM', 'TRIL_AKHEB' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL ]
  store = {
    filters = function() return weighted_store_filter(filters_PET_SHOP) end,
    max_stock = 12,
    actor_filter = function(o) return allow_sell('PET_SHOP', o) end,
  },
}

newEntity {
  define_as = 'MAYOR_S_OFFICE',
  name = "Mayor's Office",
  display = '+',
  color = colors.C_o,
  owner = resolvers.store_owner{ 'ULDRIK' },
  -- actions:
  --   Race legends (16)
  --   Town history (15)
  --   Get a quest (35)
  store = {
    max_stock = 0,
  },
}

local filters_INN = {
  { defined='RATION_OF_FOOD', id='true' },
  { defined='HARD_BISCUIT', id='true' },
  { defined='STRIP_OF_VENISON', id='true' },
  { defined='PINT_OF_FINE_WINE', id='true' },
  { defined='PINT_OF_FINE_ALE', id='true' },
}
newEntity {
  define_as = 'INN',
  name = "Inn",
  display = '+',
  color = colors.C_w,
  owner = resolvers.store_owner{ 'OTICK', 'CELEBOR' },
  -- actions:
  --   Purchase an item (2)
  --   Steal an item (4)
  --   Rest for the night (5)
  --   Buy food and drink (6)
  --   Listen for rumours (7)
  store = {
    filters = function() return weighted_store_filter(filters_INN) end,
    max_stock = 8,
  },
}

local filters_THE_SOOTHSAYER = {
  { defined='SCRL_DIVINATION', id='true', prob=20 },
  { defined='SCRL_DIVINATION', id='true', prob=20 },
  { defined='SCRL_DIVINATION', id='true', prob=20 },
  { defined='SCRL_DIVINATION', id='true', prob=20 },
  { defined='SCRL_DIVINATION', id='true', prob=20 },
  { defined='SCRL_DIVINATION', id='true', prob=20 },
}
newEntity {
  define_as = 'THE_SOOTHSAYER',
  name = "The Soothsayer",
  display = '+',
  color = colors.C_B,
  owner = resolvers.store_owner{ 'MERANA' },
  -- actions:
  --   Purchase an item (2)
  --   Presage fate (8)
  -- flags = [ RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_THE_SOOTHSAYER) end,
    max_stock = 2,
    dungeon_town = true,
  },
}

local filters_LIBRARY = {
  { defined='STAFF_IDENTIFY', id='true' },
}
newEntity {
  define_as = 'LIBRARY',
  name = "Library",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'MIRIMBAR' },
  -- actions:
  --   Research item (14)
  --   Town history (15)
  --   Race legends (16)
  --   Purchase an item (2)
  store = {
    filters = function() return weighted_store_filter(filters_LIBRARY) end,
    max_stock = 2,
  },
}

newEntity {
  define_as = 'CASTLE',
  name = "Castle",
  display = '+',
  color = colors.C_o,
  owner = resolvers.store_owner{ 'ARAGORN' },
  -- actions:
  --   Race legends (16)
  --   Get a quest (35)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'CASINO',
  name = "Casino",
  display = '+',
  color = colors.C_s,
  owner = resolvers.store_owner{ 'SONDAR' },
  -- actions:
  --   Game rules (13)
  --   In-Between (9)
  --   Play craps (10)
  --   Play dice slots (12)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'BEASTMASTER_SHANTY',
  name = "Beastmaster Shanty",
  display = '+',
  color = colors.C_g,
  owner = resolvers.store_owner{ 'SHARRA' },
  -- actions:
  --   Research monster (18)
  --   Get quest monster (21)
  --   Turn in quest corpse (22)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'FIGHTERS_HALL',
  name = "Fighters Hall",
  display = '+',
  color = colors.C_s,
  owner = resolvers.store_owner{ 'HJOLGAR' },
  -- actions:
  --   Enchant weapon (24)
  --   Enchant armour (25)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'TOWER_OF_MAGERY',
  name = "Tower of Magery",
  display = '+',
  color = colors.C_b,
  owner = resolvers.store_owner{ 'TANISTIL' },
  -- actions:
  --   Recharge item (26)
  --   Identify possessions (27)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'INNER_TEMPLE',
  name = "Inner Temple",
  display = '+',
  color = colors.C_G,
  owner = resolvers.store_owner{ 'ELDORE' },
  -- actions:
  --   Healing prayer (28)
  --   Restoration (29)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'PALADINS_GUILD',
  name = "Paladins Guild",
  display = '+',
  color = colors.C_g,
  owner = resolvers.store_owner{ 'VILIOS' },
  -- actions:
  --   Healing prayer (28)
  --   Enchant armour (25)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'RANGERS_GUILD',
  name = "Rangers Guild",
  display = '+',
  color = colors.C_u,
  owner = resolvers.store_owner{ 'ANGROS' },
  -- actions:
  --   Enchant arrows (31)
  --   Enchant bow (32)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'THUNDERLORDS_NEST',
  name = "Thunderlords' Nest",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'PALANO' },
  -- actions:
  --   Recall to dungeon (33)
  --   Purchase an item (2)
  --   Teleport to dungeon-level (34)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'THE_MIRROR',
  name = "The Mirror",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'GALADRIEL' },
  -- actions:
  --   Research item (44)
  --   Town history (15)
  --   Race legends (16)
  --   View fate (43)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'SEAT_OF_RULING',
  name = "Seat of Ruling",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'CELEBORN' },
  -- actions:
  --   Look at busts of Kings (17)
  --   Get a quest (35)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'WIZARDS_SPIRE',
  name = "Wizards Spire",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'VALCERONWE' },
  -- actions:
  --   Get a quest (60)
  --   Recharge item (26)
  --   Identify possessions (27)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'PRIESTS_CIRCLE',
  name = "Priests Circle",
  display = '+',
  color = colors.C_o,
  owner = resolvers.store_owner{ 'VORONWE' },
  -- actions:
  --   Healing prayer (28)
  --   Restoration (29)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'TOWER_OF_THE_KING',
  name = "Tower of the King",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'TURGON' },
  -- actions:
  --   Look at busts of Kings (17)
  --   Get a quest (35)
  store = {
    max_stock = 0,
  },
}

local filters_GONDOLIN_LIBRARY = {
  { defined='STAFF_IDENTIFY', id='true' },
}
newEntity {
  define_as = 'GONDOLIN_LIBRARY',
  name = "Library",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'PENGOLODH' },
  -- actions:
  --   Research item (14)
  --   Town history (15)
  --   Race legends (16)
  --   Purchase an item (2)
  store = {
    filters = function() return weighted_store_filter(filters_GONDOLIN_LIBRARY) end,
    max_stock = 2,
  },
}

local filters_THE_WHITE_TREE = {
  { defined='RATION_OF_FOOD', id='true' },
  { defined='HARD_BISCUIT', id='true' },
  { defined='STRIP_OF_VENISON', id='true' },
  { defined='PINT_OF_FINE_WINE', id='true' },
  { defined='PINT_OF_FINE_ALE', id='true' },
}
newEntity {
  define_as = 'THE_WHITE_TREE',
  name = "The White Tree",
  display = '+',
  color = colors.C_w,
  owner = resolvers.store_owner{ 'AERANDIR' },
  -- actions:
  --   Purchase an item (2)
  --   Steal an item (4)
  --   Rest for the night (5)
  --   Buy food and drink (6)
  --   Listen for rumours (7)
  store = {
    filters = function() return weighted_store_filter(filters_THE_WHITE_TREE) end,
    max_stock = 8,
  },
}

newEntity {
  define_as = 'CRAFTSMASTER',
  name = "Craftsmaster",
  display = '+',
  color = colors.C_s,
  owner = resolvers.store_owner{ 'CELEBRIMBOR' },
  -- actions:
  --   Enchant weapon (24)
  --   Enchant armour (25)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'EARTH_DOME_NATURE_',
  name = "Earth-Dome (Nature)",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'LOMELOSSE' },
  -- actions:
  --   Herbal Healing (39)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'MINSTRELS_HAVEN',
  name = "Minstrels Haven",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'ARLINDEL' },
  -- actions:
  --   Song of Lore (40)
  --   Distribute earnings (41)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'STAR_DOME',
  name = "Star-Dome",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'SULRAEN' },
  -- actions:
  --   Identify possessions (46)
  --   Recharge item (47)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'VALARIN_TEMPLE',
  name = "Valarin Temple",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'FIRIEL' },
  -- actions:
  --   Healing prayer (28)
  --   Restoration (48)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'SEA_DOME',
  name = "Sea-Dome",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'EARENDUR' },
  -- actions:
  --   Morph restoration (49)
  --   Get a quest (35)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'THE_GOLDEN_FLOWER',
  name = "The Golden Flower",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'GLORFINDEL' },
  -- actions:
  --   Enchant arrows (50)
  --   Enchant bow (51)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'THE_FOUNTAIN',
  name = "The Fountain",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'ECTHELION' },
  -- actions:
  --   Enchant armour (52)
  --   See Healers (53)
  store = {
    max_stock = 0,
  },
}

local filters_AXE_SMITH = {
  { type='weapon', subtype='axe', id='true' },
}
newEntity {
  define_as = 'AXE_SMITH',
  name = "Axe Smith",
  display = '3',
  color = colors.C_w,
  owner = resolvers.store_owner{ 'ARNOLD', 'ARNDAL', 'EDDIE', 'OGLIGN' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_AXE_SMITH) end,
    max_stock = 12,
    dungeon_town = true,
  },
}

local filters_HAFTED_SMITH = {
  { type='weapon', subtype='hafted', id='true' },
}
newEntity {
  define_as = 'HAFTED_SMITH',
  name = "Hafted Smith",
  display = '3',
  color = colors.C_w,
  owner = resolvers.store_owner{ 'ARNOLD', 'ARNDAL', 'EDDIE', 'OGLIGN' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_HAFTED_SMITH) end,
    max_stock = 12,
    dungeon_town = true,
  },
}

local filters_POLEARM_SMITH = {
  { type='weapon', subtype='polearm', id='true' },
}
newEntity {
  define_as = 'POLEARM_SMITH',
  name = "Polearm Smith",
  display = '3',
  color = colors.C_w,
  owner = resolvers.store_owner{ 'ARNOLD', 'ARNDAL', 'EDDIE', 'OGLIGN' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_POLEARM_SMITH) end,
    max_stock = 12,
    dungeon_town = true,
  },
}

local filters_SWORD_SMITH = {
  { type='weapon', subtype='sword', id='true' },
}
newEntity {
  define_as = 'SWORD_SMITH',
  name = "Sword Smith",
  display = '3',
  color = colors.C_w,
  owner = resolvers.store_owner{ 'ARNOLD', 'ARNDAL', 'EDDIE', 'OGLIGN' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_SWORD_SMITH) end,
    max_stock = 12,
    dungeon_town = true,
  },
}

local filters_RARE_JEWELRY_SHOP = {
  { type='jewelry', subtype='amulet', id='true' },
  { type='jewelry', subtype='ring', id='true' },
}
newEntity {
  define_as = 'RARE_JEWELRY_SHOP',
  name = "Rare Jewelry Shop",
  display = '6',
  color = colors.C_v,
  owner = resolvers.store_owner{ 'ARIEL', 'BUGGERBY', 'INGLORIAN', 'LUTHIEN' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ VERY_RARE DEEP_LEVEL FORCE_LEVEL DEPEND_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_RARE_JEWELRY_SHOP) end,
    max_stock = 10,
    dungeon_town = true,
  },
}

local filters_JEWELRY_SHOP = {
  { type='jewelry', subtype='amulet', id='true' },
  { type='jewelry', subtype='ring', id='true' },
}
newEntity {
  define_as = 'JEWELRY_SHOP',
  name = "Jewelry Shop",
  display = '6',
  color = colors.C_y,
  owner = resolvers.store_owner{ 'ARIEL', 'BUGGERBY', 'INGLORIAN', 'LUTHIEN' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ RARE MEDIUM_LEVEL FORCE_LEVEL DEPEND_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_JEWELRY_SHOP) end,
    max_stock = 20,
    dungeon_town = true,
  },
}

local filters_FOOTWEAR_SHOP = {
  { type='armor', subtype='footwear', id='true' },
}
newEntity {
  define_as = 'FOOTWEAR_SHOP',
  name = "Footwear Shop",
  display = '2',
  color = colors.C_r,
  owner = resolvers.store_owner{ 'KON_DAR', 'DARG_LOW', 'DECADO', 'WIELAND' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL COMMON RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_FOOTWEAR_SHOP) end,
    max_stock = 12,
    dungeon_town = true,
  },
}

local filters_RARE_FOOTWEAR_SHOP = {
  { type='armor', subtype='footwear', id='true' },
}
newEntity {
  define_as = 'RARE_FOOTWEAR_SHOP',
  name = "Rare Footwear Shop",
  display = '2',
  color = colors.C_r,
  owner = resolvers.store_owner{ 'KON_DAR', 'DARG_LOW', 'DECADO', 'WIELAND' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ VERY_RARE MEDIUM_LEVEL DEEP_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_RARE_FOOTWEAR_SHOP) end,
    max_stock = 8,
    dungeon_town = true,
  },
}

local filters_DGN_TOWN_LIBRARY = {
  { type='magic', subtype='spellbook', id='true' },
  { properties={ 'is_demongear', id='true' } },
}
newEntity {
  define_as = 'DGN_TOWN_LIBRARY',
  name = "Library",
  display = '9',
  color = colors.C_y,
  owner = resolvers.store_owner{ 'KON_DAR', 'DARG_LOW', 'DECADO', 'WIELAND' },
  -- actions:
  --   Identify possessions (27)
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL DEPEND_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_DGN_TOWN_LIBRARY) end,
    max_stock = 24,
    dungeon_town = true,
  },
}

local filters_FORBIDDEN_LIBRARY = {
  { type='magic', subtype='spellbook', id='true' },
  { properties={ 'is_demongear', id='true' } },
}
newEntity {
  define_as = 'FORBIDDEN_LIBRARY',
  name = "Forbidden Library",
  display = '9',
  color = colors.C_v,
  owner = resolvers.store_owner{ 'KON_DAR', 'DARG_LOW', 'DECADO', 'WIELAND' },
  -- actions:
  --   Identify possessions (27)
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ RARE DEEP_LEVEL DEPEND_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_FORBIDDEN_LIBRARY) end,
    max_stock = 12,
    dungeon_town = true,
  },
}

newEntity {
  define_as = 'EXPENSIVE_BLACK_MARKET',
  name = "Expensive Black Market",
  display = '7',
  color = colors.C_v,
  owner = resolvers.store_owner{ 'GARY', 'HISTOR', 'ZORAK', 'TIPO' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL DEEP_LEVEL DEPEND_LEVEL ALL_ITEM RANDOM ]
  store = {
    filters = black_market_filter,
    max_stock = 12,
    dungeon_town = true,
  },
}

local filters_COMMON_SHOP = {
  { defined='BRASS_LANTERN', id='true', prob=95 },
  { defined='FLASK_OF_OIL', id='true' },
  { defined='RATION_OF_FOOD', id='true' },
  { defined='SHOVEL', id='true', prob=60 },
  { defined='PICK', id='true', prob=50 },
  { defined='IRON_SPIKE', id='true' },
  { defined='IRON_SHOT', id='true', prob=70 },
  { defined='BOLT', id='true', prob=70 },
  { defined='ARROW', id='true', prob=70 },
  { defined='CLOAK', id='true', prob=98 },
  { defined='FUR_CLOAK', id='true', prob=46 },
  { defined='SCRL_WORD_OF_RECALL', id='true' },
  { defined='POT_CURE_CRITICAL_WOUNDS', id='true' },
}
newEntity {
  define_as = 'COMMON_SHOP',
  name = "Common Shop",
  display = '1',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'BILBO', 'RAISTLIN', 'SULTAN', 'LYAR_EL' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_COMMON_SHOP) end,
    max_stock = 12,
    dungeon_town = true,
  },
}

local filters_DRAGON_HUNTER = {
  { properties={ 'is_dragon_armor', id='true' } },
}
newEntity {
  define_as = 'DRAGON_HUNTER',
  name = "Dragon Hunter",
  display = '2',
  color = colors.C_v,
  owner = resolvers.store_owner{ 'KON_DAR', 'DARG_LOW', 'DECADO', 'WIELAND' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ VERY_RARE DEEP_LEVEL DEPEND_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_DRAGON_HUNTER) end,
    max_stock = 12,
    dungeon_town = true,
  },
}

local filters_SPEED_RING_MARKET = {
  { defined='RING_SPEED', id='true' },
}
newEntity {
  define_as = 'SPEED_RING_MARKET',
  name = "Speed Ring Market",
  display = '6',
  color = colors.C_G,
  owner = resolvers.store_owner{ 'ARIEL', 'BUGGERBY', 'INGLORIAN', 'LUTHIEN' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ VERY_RARE DEPEND_LEVEL SHALLOW_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_SPEED_RING_MARKET) end,
    max_stock = 6,
    dungeon_town = true,
  },
}

local filters_SCRIBE = {
  { type='utility', subtype='scroll', id='true' },
}
newEntity {
  define_as = 'SCRIBE',
  name = "Scribe",
  display = '5',
  color = colors.C_B,
  owner = resolvers.store_owner{ 'KON_DAR', 'DARG_LOW', 'DECADO', 'WIELAND' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL DEPEND_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_SCRIBE) end,
    max_stock = 12,
    dungeon_town = true,
  },
}

local filters_POTION_STORE = {
  { type='utility', subtype='potion', id='true' },
  { type='utility', subtype='potion', id='true' },
}
newEntity {
  define_as = 'POTION_STORE',
  name = "Potion Store",
  display = '5',
  color = colors.C_B,
  owner = resolvers.store_owner{ 'KON_DAR', 'DARG_LOW', 'DECADO', 'WIELAND' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL DEPEND_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_POTION_STORE) end,
    max_stock = 12,
    dungeon_town = true,
  },
}

local filters_RECALLER = {
  { defined='SCRL_WORD_OF_RECALL', id='true' },
}
newEntity {
  define_as = 'RECALLER',
  name = "Recaller",
  display = '+',
  color = colors.C_b,
  owner = resolvers.store_owner{ 'KON_DAR', 'DARG_LOW', 'DECADO', 'WIELAND' },
  -- actions:
  --   Recall to dungeon (33)
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  -- flags = [ COMMON RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_RECALLER) end,
    max_stock = 2,
    dungeon_town = true,
  },
}

local filters_MASTER_ARCHER = {
  { defined='SLING', id='true' },
  { defined='SHORT_BOW', id='true' },
  { defined='LONG_BOW', id='true', prob=20 },
  { defined='LIGHT_CROSSBOW', id='true' },
  { defined='HEAVY_CROSSBOW', id='true', prob=20 },
  { type='ammo', subtype='shots', id='true', prob=50 },
  { type='ammo', subtype='arrows', id='true', prob=50 },
  { type='ammo', subtype='bolts', id='true', prob=50 },
}
newEntity {
  define_as = 'MASTER_ARCHER',
  name = "Master Archer",
  display = '3',
  color = colors.C_g,
  owner = resolvers.store_owner{ 'ARNOLD', 'ARNDAL', 'EDDIE', 'OGLIGN' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ RARE MEDIUM_LEVEL DEPEND_LEVEL RANDOM ]
  store = {
    filters = function() return weighted_store_filter(filters_MASTER_ARCHER) end,
    max_stock = 24,
    dungeon_town = true,
  },
}

newEntity {
  define_as = 'MERCHANTS_GUILD',
  name = "Merchants Guild",
  display = '+',
  color = colors.C_g,
  owner = resolvers.store_owner{ 'KANRIS' },
  -- actions:
  --   Request an item (56)
  --   Ask for loan (57)
  --   Pay back loan (58)
  store = {
    max_stock = 0,
  },
}

newEntity {
  define_as = 'THE_MATHOM_HOUSE',
  name = "The Mathom-house",
  display = '+',
  color = colors.C_g,
  owner = resolvers.store_owner{ 'BILBO' },
  -- actions:
  --   Donate an item (59)
  --   Examine an item (3)
  -- flags = [ MUSEUM ]
  store = {
    max_stock = 255,
    mathom_house = true,
  },
}

local filters_THE_PRANCING_PONY = {
  { defined='RATION_OF_FOOD', id='true' },
  { defined='HARD_BISCUIT', id='true' },
  { defined='STRIP_OF_VENISON', id='true' },
  { defined='PINT_OF_FINE_WINE', id='true' },
  { defined='PINT_OF_FINE_ALE', id='true' },
}
newEntity {
  define_as = 'THE_PRANCING_PONY',
  name = "The Prancing Pony",
  display = '+',
  color = colors.C_w,
  owner = resolvers.store_owner{ 'BARLIMAN' },
  -- actions:
  --   Purchase an item (2)
  --   Steal an item (4)
  --   Rest for the night (5)
  --   Buy food and drink (6)
  --   Listen for rumours (7)
  store = {
    filters = function() return weighted_store_filter(filters_THE_PRANCING_PONY) end,
    max_stock = 8,
  },
}

local filters_MINING_SUPPLY_STORE = {
  { type='tool', subtype='digger', id='true' },
  { defined='WOODEN_TORCH', id='true' },
  { defined='BRASS_LANTERN', id='true', prob=95 },
  { defined='FLASK_OF_OIL', id='true', prob=95 },
  { defined='DWARVEN_LANTERN', id='true', prob=75 },
  { defined='FEANORIAN_LAMP', id='true', prob=60 },
  { defined='WAND_TIDAL_WAVE', id='true', prob=60 },
  { defined='POT_DETONATIONS', id='true', prob=70 },
}
newEntity {
  define_as = 'MINING_SUPPLY_STORE',
  name = "Mining Supply store",
  display = '+',
  color = colors.C_s,
  owner = resolvers.store_owner{ 'WIELAND', 'OGLIGN', 'LUDWIG', 'ZORAK' },
  -- actions:
  --   Sell an item (1)
  --   Purchase an item (2)
  --   Examine an item (3)
  --   Steal an item (4)
  -- flags = [ MEDIUM_LEVEL ]
  store = {
    filters = function() return weighted_store_filter(filters_MINING_SUPPLY_STORE) end,
    max_stock = 24,
    actor_filter = function(o) return allow_sell('MINING_SUPPLY_STORE', o) end,
  },
}

local filters_MINAS_ANOR_LIBRARY = {
  { defined='STAFF_IDENTIFY', id='true' },
}
newEntity {
  define_as = 'MINAS_ANOR_LIBRARY',
  name = "Library",
  display = '+',
  color = colors.C_U,
  owner = resolvers.store_owner{ 'MIRIMBAR' },
  -- actions:
  --   Get a quest (61)
  --   Research item (14)
  --   Town history (15)
  --   Race legends (16)
  --   Purchase an item (2)
  store = {
    filters = function() return weighted_store_filter(filters_MINAS_ANOR_LIBRARY) end,
    max_stock = 2,
  },
}
