-- $Id: Player.lua 79 2012-09-20 04:24:03Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require "engine.class"
require "mod.class.Actor"
require 'mod.class.interface.PlayerCombat'
require 'mod.class.interface.PlayerSkills'
require 'mod.class.interface.PlayerFates'
require "engine.interface.PlayerRest"
require "engine.interface.PlayerRun"
require "engine.interface.PlayerMouse"
require "engine.interface.PlayerHotkeys"
local Map = require "engine.Map"
local Dialog = require "engine.Dialog"
local DeathDialog = require "mod.dialogs.DeathDialog"
local TacticExplorDialog = require 'mod.dialogs.TacticExplorDialog'
local Astar = require"engine.Astar"
local DirectPath = require "engine.DirectPath"

--- Defines the player
-- It is a normal actor, with some redefined methods to handle user interaction.<br/>
-- It is also able to run and rest and use hotkeys
module(..., package.seeall, class.inherit(mod.class.Actor,
					  mod.class.interface.PlayerCombat,
					  mod.class.interface.PlayerSkills,
					  mod.class.interface.PlayerFates,
					  engine.interface.PlayerRest,
					  engine.interface.PlayerRun,
					  engine.interface.PlayerMouse,
					  engine.interface.PlayerHotkeys))

local exp_chart = {
  [1]	= 0,
  [2]	= 10,
  [3]	= 25,
  [4]	= 45,
  [5]	= 70,
  [6]	= 100,
  [7]	= 140,
  [8]	= 200,
  [9]	= 280,
  [10]	= 380,
  [11]	= 500,
  [12]	= 650,
  [13]	= 850,
  [14]	= 1100,
  [15]	= 1400,
  [16]	= 1800,
  [17]	= 2300,
  [18]	= 2900,
  [19]	= 3600,
  [20]	= 4400,
  [21]	= 5400,
  [22]	= 6800,
  [23]	= 8400,
  [24]	= 10200,
  [25]	= 12500,
  [26]	= 17500,
  [27]	= 25000,
  [28]	= 35000,
  [29]	= 50000,
  [30]	= 75000,
  [31]	= 100000,
  [32]	= 150000,
  [33]	= 200000,
  [34]	= 275000,
  [35]	= 350000,
  [36]	= 450000,
  [37]	= 550000,
  [38]	= 700000,
  [39]	= 850000,
  [40]	= 1000000,
  [41]	= 1250000,
  [42]	= 1500000,
  [43]	= 1800000,
  [44]	= 2100000,
  [45]	= 2400000,
  [46]	= 2700000,
  [47]	= 3000000,
  [48]	= 3500000,
  [49]	= 4000000,
  [50]	= 4500000,
}

function _M:init(t, no_default)
  t.display=t.display or '@'
  t.color_r=t.color_r or 230
  t.color_g=t.color_g or 230
  t.color_b=t.color_b or 230

  t.player = true
  t.type = t.type or "humanoid"
  t.subtype = t.subtype or "player"
  t.faction = t.faction or "players"

  t.lite = t.lite or 0

  mod.class.Actor.init(self, t, no_default)
  mod.class.interface.PlayerCombat.init(self, t)
  mod.class.interface.PlayerSkills.init(self, t)
  mod.class.interface.PlayerFates.init(self, t)
  engine.interface.PlayerHotkeys.init(self, t)

  self.descriptor = {}
  self.max_level = 50
  self.max_exp = self.exp
  self.level_hiwater = self.level
  self.speed = 0
end

function _M:getExpChart(level)
  if not level or level < 1 then return 0 end
  return math.floor((exp_chart[level] * (self.xp_mult or 100))/100)
end

function _M:move(x, y, force)
  local moved = mod.class.Actor.move(self, x, y, force)
  if moved then
    game.level.map:moveViewSurround(self.x, self.y, 8, 8)

    self:describeFloor(self.x, self.y)
  end

  -- Remember not to describe this grid again.
  self.old_x, self.old_y = self.x, self.y

  return moved
end

function _M:describeFloor(x, y)
  if self.old_x == self.x and self.old_y == self.y then return end

  -- Auto-pickup stuff from floor.
  local i = 1
  local obj = game.level.map:getObject(x, y, i)
  while obj do
    if obj.auto_pickup and self:pickupFloor(i, true) then
      -- Nothing to do.
    else
      i = i + 1
    end
    obj = game.level.map:getObject(x, y, i)
  end

  -- i is now one higher than the number of objects on the floor.
  -- TODO Prompt to pickup, probably controlled by a setting.
  if i == 2 then
    game.log('On floor:  %s', game.level.map:getObject(x, y, 1):getName())
  elseif i > 2 then
    game.log('There are %d objects here.', i - 1)
  end
end

function _M:act()
  if not mod.class.Actor.act(self) then return end

  -- Clean log flasher
  game.flash:empty()

  -- Resting ? Running ? Otherwise pause
  if not self:restStep() and not self:runStep() and self.player then
    game.paused = true
  end
end

-- Precompute FOV form, for speed
local fovdist = {}
for i = 0, 30 * 30 do
  fovdist[i] = math.max((20 - math.sqrt(i)) / 14, 0.6)
end

function _M:playerFOV()
  -- Clean FOV before computing it
  game.level.map:cleanFOV()
  -- Compute both the normal and the lite FOV, using cache
  self:computeFOV(self.sight or 20, "block_sight", function(x, y, dx, dy, sqdist)
    game.level.map:apply(x, y, fovdist[sqdist])
  end, true, false, true)
  self:computeFOV(self.lite, "block_sight", function(x, y, dx, dy, sqdist) game.level.map:applyLite(x, y) end, true, true, true)
end

--- Called before taking a hit, overload mod.class.Actor:onTakeHit() to stop resting and running
function _M:onTakeHit(value, src)
  self:runStop("taken damage")
  self:restStop("taken damage")
  local ret = mod.class.Actor.onTakeHit(self, value, src)
  if self.life < self.max_life * 0.3 then
    local sx, sy = game.level.map:getTileToScreen(self.x, self.y)
    game.flyers:add(sx, sy, 30, (rng.range(0,2)-1) * 0.5, 2, "LOW HEALTH!", {255,0,0}, true)
  end
  return ret
end

function _M:die(src)
  if self.game_ender then
    engine.interface.ActorLife.die(self, src)
    game.paused = true
    self.energy.value = game.energy_to_act
    game:registerDialog(DeathDialog.new(self))
  else
    mod.class.Actor.die(self, src)
  end
end

function _M:setName(name)
  self.name = name
  game.save_name = name
end

--- Notify the player of available cooldowns
function _M:onTalentCooledDown(tid)
  local t = self:getTalentFromId(tid)

  local x, y = game.level.map:getTileToScreen(self.x, self.y)
  game.flyers:add(x, y, 30, -0.3, -3.5, ("%s available"):format(t.name:capitalize()), {0,255,00})
  game.log("#00ff00#Talent %s is ready to use.", t.name)
end

function _M:levelup()
  local newlev = self:levelupMsg()

  self:recalcEverything()

  if newlev then
    self.skill_points = self.skill_points + 6
    game.log("#00FFFF#%s has %d skill points to spend.  Press 'G' to use them.", self.name, self.skill_points)
  end
end

--- Tries to get a target from the user
function _M:getTarget(typ)
  return game:targetGetForPlayer(typ)
end

--- Sets the current target
function _M:setTarget(target)
  return game:targetSetForPlayer(target)
end

local function spotHostiles(self)
  local seen = false
  -- Check for visible monsters, only see LOS actors, so telepathy wont prevent resting
  core.fov.calc_circle(self.x, self.y, game.level.map.w, game.level.map.h, 20, function(_, x, y) return game.level.map:opaque(x, y) end, function(_, x, y)
  local actor = game.level.map(x, y, game.level.map.ACTOR)
  if actor and self:reactionToward(actor) < 0 and self:canSee(actor) and game.level.map.seens(x, y) then seen = true end
end, nil)
return seen
end

--- Can we continue resting ?
-- We can rest if no hostiles are in sight, and if we need life/mana/stamina (and their regen rates allows them to fully regen)
function _M:restCheck()
  if spotHostiles(self) then return false, "hostile spotted" end

  if self.resting.rest_turns and self.repeat_action then
    -- Slight Hack(TM):  We re-use the resting infrastructure to repeat
    -- certain actions like attempting to open locked doors.
    return self:repeat_action()
  end

  -- Check resources, make sure they CAN go up, otherwise we will never stop
  --[[
  if self:getPower() < self:getMaxPower() and self.power_regen > 0 then return true end
  ]]
  if self.life < self.max_life and self.life_regen > 0 then return true end

  return false, "all resources and life at maximum"
end

function _M:onRestStop()
  -- Remove any repeating action.
  self.repeat_action = nil
end

--- Can we continue running?
-- We can run if no hostiles are in sight, and if we no interesting terrains are next to us
function _M:runCheck()
  if spotHostiles(self) then return false, "hostile spotted" end

  -- Notice any noticeable terrain
  local noticed = false
  self:runScan(function(x, y)
    -- Only notice interesting terrains
    local grid = game.level.map(x, y, Map.TERRAIN)
    if grid and grid.notice then noticed = "interesting terrain" end
  end)
  if noticed then return false, noticed end

  self:playerFOV()

  return engine.interface.PlayerRun.runCheck(self)
end

--- Move with the mouse
-- We just feed our spotHostile to the interface mouseMove
function _M:mouseMove(tmx, tmy)
  return engine.interface.PlayerMouse.mouseMove(self, tmx, tmy, spotHostiles)
end

function _M:playerOpen()
  local count = 0
  local tx, ty
  for x = self.x - 1, self.x + 1 do
    for y = self.y - 1, self.y + 1 do
      if x ~= self.x or y ~= self.y then
	local c = game.level.map:checkAllEntitiesCount(x, y, 'is_openable')
	count = count + c
	if c == 1 then tx, ty = x, y end
      end
    end
  end
  if count == 0 then
    game.log('You see nothing here to open.')
    return
  elseif count > 1 then
    game.log('Which direction?')
    tx, ty = self:getTarget{ range=1, immediate_keys=true, default_target=self.player }
    if not tx or not ty then return end
  end
  -- tx, ty should now be the location of what we want to open.
  local g = game.level.map(tx, ty, Map.TERRAIN)
  local a = game.level.map(tx, ty, Map.ACTOR)
  local oo = game.level.map:checkAllEntitiesNoStop(x, y, 'on_open_chest')
  -- Slight Hack(TM):  Get one of the chests out of oo.
  local o = nil
  for _, o_ in pairs(oo) do o = o_ break end
  if not g.door_opened and not o then
    game.log('You see nothing there to open.')
    return
  elseif a then
    game.log('There is a monster in the way!')
    -- Slight Hack(TM):  Move in the direction of the monster to attack it.
    local dir = (self.x - tx + 1)*3 + ty - self.ty + 2
    self.moveDir(dir)
  elseif o then
    -- Slight Hack(TM):  We hijack the resting infrastructure to try this
    -- multiple times.
    self.repeat_action = function(self)
    end
    self:restInit(99, 'attempting to open chest', 'attempted to open chest')
  elseif g and g.door_opened then
    if not g.try_open_door then
      -- Door is not locked or jammed; just open it.
      g:openDoor(tx, ty, self)
      self:useEnergy()
      return
    end
    -- Slight Hack(TM):  As above.
    self.repeat_action = function(self)
      local success, can_retry = g:try_open_door(self)
      if success then
	g:openDoor(tx, ty, self)
	return false, nil
      end
      return can_retry, msg
    end
    self:restInit(99, 'attempting to unlock door', 'attempted to unlock door')
  end
end

function _M:playerWear()
  local inven = self:getInven('INVEN')
  local filter_cb = function(o)
    return o:wornInven() and self:getInven(o:wornInven()) and true or false
  end
  local action_cb = function(o, item)
    self:doWear(inven, item, o)
    return true
  end
  self:showInventory('Wear/Wield', inven, filter_cb, action_cb)
end

function _M:playerTakeoff()
  local action_cb = function(ok, inven, item)
    self:doTakeoff(inven, item, o)
    return true
  end
  self:showEquipment('Take Off', nil, action_cb)
end

function _M:playerPickup()
  if game.level.map:getObject(self.x, self.y, 2) then
    local d
    local action = function(o, item)
      self:pickupFloor(item, true)
      self.changed = true
      d:used()
    end
    d = self:showPickupFloor('Pickup', nil, action)
  else
    self:pickupFloor(1, true)
    self:useEnergy()
    self.changed = true
  end
end

function _M:playerDrop()
  local inven = self:getInven(self.INVEN_INVEN)
  local action = function(o, item)
    self:dropFloor(inven, item, true, true)
    self:sortInven(inven)
    self:useEnergy()
    self.changed = true
    return true
  end
  self:showInventory('Drop', inven, nil, action)
end

function _M:doWear(inven, item, o)
  self:removeObject(inven, item, true)
  local ro = self:wearObject(o, true, true)
  if ro then
    if type(ro) == "table" then self:addObject(inven, ro) end
  elseif not ro then
    self:addObject(inven, o)
  end
  self:sortInven()
  self:recalcEverything()
  self:useEnergy()
  self.changed = true
end

function _M:doTakeoff(inven, item, o)
  if self:takeoffObject(inven, item) then
    self:addObject(self.INVEN_INVEN, o)
  end
  self:sortInven()
  self:useEnergy()
  self.changed = true
end

function _M:doDrop(inven, item, on_done)
  self:dropFloor(inven, item, true, true)
  self:sortInven(inven)
  self:recalcEverything()
  self:useEnergy()
  self.changed = true
  if on_done then on_done() end
end

local super_gainExp = _M.gainExp
function _M:gainExp(value)
  -- If our exp has been drained, increase max_exp by a fraction of what
  -- we've gained.
  if self.exp < self.max_exp then self.max_exp = self.max_exp + value/5 end

  -- TODO Siphon off some of the exp to sentient weapons and/or Alchemist
  -- artifacts.
  super_gainExp(self, value)

  self.max_exp = math.max(self.max_exp, self.exp)
end

function _M:getLuckScale(min, max)
  local luck = util.bound(self.luck, -30, 30)
  return min + math.floor((luck + 30)*(max - min)/60)
end

function _M:recalcHP()
  self.max_life = self.base_hp_by_lev[self.level] +
		  math.floor(self:getStatRelated('con_maxhp')*self.level/2)
  self.max_life = math.max(self.max_life, self.level + 1)
  self.life = math.min(self.life, self.max_life)
end

function _M:recalcMana()
  local mana = self:getSkillLevel('MAGIC', 200) +
	       math.floor(self:getStatRelated('int_wis_mana')*self.level/4)
  if mana > 0 then mana = mana + 1 end
  mana = math.floor(mana * self.mana_mult)

  local aw = self:armorWeight()
  local sk_c = self:getSkillLevel('COMBAT')
  mana = mana - math.max(0, math.floor(aw - sk_c -  20))

  self.max_mana = math.max(mana, 0)
  self.mana = math.min(self.mana, self.max_mana)
end

function _M:recalcEquip()
  -- TODO Add light modifiers from birth descriptors.
  for slot = 1, #self.inven_def do
    if self.inven[slot] and self.inven_def[slot].is_worn then
      for _, o in ipairs(self.inven[slot]) do
	self.lite = self.lite + (o.lite or 0)
	self.ac = self.ac + (o.armor or 0)
      end
    end
  end
end

function _M:applyLevelGains()
  for lev, t in ipairs(self.gain_at_level) do
    if lev <= self.level then
      for _, ab in pairs(t.abilities or {}) do
	self.abilities[ab] = true
      end
      for stat, delta in pairs(t.stats or {}) do
	self.stats[stat].equip_adj = self.stats[stat].equip_adj + delta
      end
      -- TODO apply mods and flags
    end
  end
end

local to_clear = {
  'speed', 'lite', 'ac',
}
function _M:recalcEverything()
  print('[PLAYER] recalcEverything')
  for _, def in pairs(self.skills_def) do
    if def.player_calc then def.player_calc(self, true) end
  end
  for _, spec in pairs(self.stats) do spec.equip_adj = 0 end
  for _, f in pairs(to_clear) do self[f] = 0 end
  self.speed = 0
  self.abilities = table.clone(self.bought_abilities)

  self:recalcEquip()
  self:applyLevelGains()

  for _, def in pairs(self.skills_def) do
    if def.player_calc then def.player_calc(self) end
  end

  TacticExplorDialog:apply(self)

  self.to_h = self.to_h + self:getStatRelated('dex_to_hit') +
			  self:getStatRelated('str_to_hit')

  self.stealth_exp = 2 ^ (30 - util.bound(self.skill_stealth, 0, 30))
  print('[PLAYER] recalcEverything:  stealth ', self.skill_stealth, ' -> ', self.stealth_exp)

  self:computeGlobalSpeed()
  self:recalcHP()
  self:recalcMana()

  self:playerFOV()

  self.changed = true
  print('[PLAYER] recalcEverything:  done')
end

-- TODO Maybe some major display, a la T4's BigNews
function _M:on_quest_grant(quest)
  game.logPlayer(game.player, "#LIGHT_GREEN#Accepted quest '%s'! #WHITE#(Press 'j' to see the quest log)", quest.name)
end

function _M:on_quest_status(quest, status, sub)
  if sub then
    game.logPlayer(game.player, "#LIGHT_GREEN#Quest '%s' status updated! #WHITE#(Press 'j' to see the quest log)", quest.name)
  elseif status == engine.Quest.COMPLETED then
    game.logPlayer(game.player, "#LIGHT_GREEN#Quest '%s' completed! #WHITE#(Press 'j' to see the quest log)", quest.name)
  elseif status == engine.Quest.DONE then
    game.logPlayer(game.player, "#LIGHT_GREEN#Quest '%s' is done! #WHITE#(Press 'j' to see the quest log)", quest.name)
  elseif status == engine.Quest.FAILED then
    game.logPlayer(game.player, "#LIGHT_RED#Quest '%s' is failed! #WHITE#(Press 'j' to see the quest log)", quest.name)
  end
end
