-- $Id: StatsDisplay.lua 8 2012-07-10 02:24:59Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'
local ActorStats = require 'mod.class.interface.ActorStats'

module(..., package.seeall, class.make)

function _M:init(actor, x, y, w, h, bgcolor, fontname, fontsize)
  self.actor = actor
  self.bgcolor = bgcolor or { 0, 0, 0 }
  self.font = core.display.newFont(fontname or '/data/font/DroidSansMono.ttf', fontsize or 10)
  self.font_h = self.font:lineSkip()
  self:resize(x, y, w, h)
end

function _M:resize(x, y, w, h)
  self.display_x, self.display_y = math.floor(x), math.floor(y)
  self.w, self.h = math.floor(w), math.floor(h)
  self.surface = core.display.newSurface(w, h)
  self.texture, self.texture_w, self.texture_h = self.surface:glTexture()
  if self.actor then
    self.actor.changed = true
  end

  local cw, ch = self.font:size(' ')
  self.font_w = cw
end

function _M:display()
  if not self.actor or not self.actor.changed or self.actor.creating_player then
    return self.surface
  end
  self.surface:erase(self.bgcolor[1], self.bgcolor[2], self.bgcolor[3])

  local aa = self.actor
  local y = 5
  for _, stat in ipairs(ActorStats.stats) do
    local c = aa:isStatMaxed(stat) and 'LIGHT_BLUE' or
	      aa:isStatDrained(stat) and 'YELLOW' or 'LIGHT_GREEN'
    local stat_fmt = aa:getStat(stat, { fmt='default' })
    stat_fmt = '#'..c..'#'..stat..': '..stat_fmt..'#LAST#'
    self.surface:drawColorStringBlended(self.font, stat_fmt, 5, y, 255, 255, 255)
    y = y + self.font_h
  end

  y = y + self.font_h
  y = self:drawResource(y, 'HP:   ', aa.life, aa.max_life)
  if aa.max_mana then
    y = self:drawResource(y, 'Mana: ', aa.mana, aa.max_mana)
  end
  if aa.gold then
    local str = ('#GOLD#Gold: %d#LAST#'):format(aa.gold)
    self.surface:drawColorStringBlended(self.font, str, 5, y, 255, 255, 255)
    y = y + self.font_h
  end
  if aa.god then
    local c = aa.praying and '#LIGHT_BLUE#' or aa.piety < 0 and '#YELLOW#' or '#LIGHT_GREEN#'
    local str = c..'Piety:  '..aa.piety..'#LAST#'
    self.surface:drawColorStringBlended(self.font, str, 5, y, 255, 255, 255)
    y = y + self.font_h
  end

  local spd_desc = aa.speed > 0 and 'Fast' or aa.speed < 0 and 'Slow' or 'Normal'
  local c_spd = aa.speed > 0 and '#00FF00#' or aa.speed < 0 and '#YELLOW#' or '#SLATE#'
  local spd_str = (c_spd..'Speed:  '..spd_desc..' (%+d)'):format(aa.speed)
  self.surface:drawColorStringBlended(self.font, spd_str, 5, y, 255, 255, 255)
  y = y + self.font_h

  -- TODO Find a better place to put this.
  local c_lev = aa.level < aa.level_hiwater and '#YELLOW#' or '#LIGHT_GREEN#'
  local lev_str = ("Level:  %d"):format(aa.level)
  lev_str = c_lev..lev_str..'#LAST#'
  self.surface:drawColorStringBlended(self.font, lev_str, 5, y, 255, 255, 255)
  y = y + self.font_h

  local c_exp = aa.exp < aa.max_exp and '#YELLOW#' or '#LIGHT_GREEN#'
  if aa.level >= aa.max_level then
    local str = c_exp..'Exp:    ********#LAST#'
    self.surface:drawColorStringBlended(self.font, str, 5, y, 255, 255, 255)
  else
    local mode = config.settings.tome2.exp_display
    local xlev = aa:getExpChart(aa.level)
    local xlevp1 = aa:getExpChart(aa.level+1)
    local str
    if mode == 'Absolute' then
      str = string.format('Exp:    %0.2f', aa.exp)
    elseif mode == 'Percent' then
      local pct = math.floor((aa.exp - xlev)/(xlevp1 - xlev) * 100)
      str = string.format('Exp:    %d', pct)
    else
      str = string.format('Exp:    %.2f', xlevp1 - aa.exp)
    end
    str = c_exp..str..'#LAST#'
    self.surface:drawColorStringBlended(self.font, str, 5, y, 255, 255, 255)
  end
  y = y + self.font_h

  local symb = aa.getInven('SYMBIOTE')
  if symb and symb[1] and symb[1].mon and not symb[1].mon.dead then
    y = self:drawResource(y, 'SymHP:', symb[1].mon.life, symb[1].mon.max_life)
  end

  -- TODO Find a better place to put this.
  if game.zone and game.level and game.level.level then
    y = y + self.font_h
    local str = game.zone.name
    if game.zone.max_level > 1 then
      str = ("%s (%d)"):format(str, game.level.level)
    end
    str = '#0000FF#'..str..'#LAST#'
    self.surface:drawColorStringBlended(self.font, str, 5, y, 255, 255, 255)
    y = y + self.font_h
  end

  -- TODO 'DTrap' if traps have been detected in the player's grid.

  for eff, p in pairs(aa.tmp) do
    local def = aa.tempeffect_def[eff]
    local c = def.status == 'beneficial' and 'LIGHT_GREEN' or 'RED'
    local str = '#'..c..'#'..def.desc..' ('..p.dur..')#LAST#'
    self.surface:drawColorStringBlended(self.font, str, 5, y, 255, 255, 255)
    y = y + self.font_h
  end

  self.surface:updateTexture(self.texture)
  return self.surface
end

function _M:drawResource(y, pfx, amt, max_amt)
  -- TODO Make this hard-coded 50% threshold configurable.
  local c = amt < max_amt/2 and 'RED' or
	    amt < max_amt and 'YELLOW' or 'LIGHT_GREEN'
  local str = '#'..c..'#'..pfx..'  '..amt..'/'..max_amt..'#LAST#'
  self.surface:drawColorStringBlended(self.font, str, 5, y, 255, 255, 255)
  return y + self.font_h
end

function _M:toScreen()
  self:display()
  self.texture:toScreenFull(self.display_x, self.display_y, self.w, self.h, self.texture_w, self.texture_h)
end
