-- $Id: Store.lua 82 2012-09-20 04:30:24Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'
local Store = require 'engine.Store'
local Dialog = require 'engine.ui.Dialog'

module(..., package.seeall, class.inherit(Store))

_M.stores_def = {}
_M.stock_levels = {
  -- NOTE:  armor/light doesn't collide with equipment/light because we
  -- check type before subtype.
  oil = 'food',
  light = 'food',
  food = {
    { cost=5, dice=3, sides=5 },
    { cost=20, dice=3, sides=5 },
  },
  scroll = 'potion',
  potion = {
    { cost=60, dice=3, sides=5 },
    { cost=240, dice=1, sides=5 },
  },
  spellbook = {
    { cost=50, dice=2, sides=3 },
    { cost=500, dice=1, sides=3 },
  },
  weapon = 'armor',
  shooter = 'armor',
  armor = {
    { cost=10, dice=3, sides=5 },
    { cost=100, dice=3, sides=5 },
  },
  spike = 'ammo',
  ammo = {
    { cost=5, dice=5, sides=5 },
    { cost=50, dice=5, sides=5 },
    { cost=500, dice=5, sides=5 },
  },
  staff = 'wand',
  wand = {
    { cost=1600, dice=1, sides=5 },
    { cost=3200, dice=1, sides=3 },
  },
}

function _M:loadStores(f)
  self.stores_def = self:loadList(f)
end

function _M:init(t, no_default)
  t.store.min_fill = t.store.min_fill or 6
  t.store.max_fill = t.store.max_fill or 24
  t.store.post_filter = t.store.post_filter or self.stockPostFilter
  t.store.purse = t.store.purse or 2500
  -- 1000 player turns at normal speed.
  t.store.restock_after = t.restock_after or 10000
  Store.init(self, t, no_default)
  -- There's no obvious way to override the 't.body = {INVEN=10000}' in
  -- Store.init(), so we patch that up here.
  self:getInven('INVEN').max = t.max_stock or 24
  -- Can't sell to stores in dungeon towns.
  if t.store.dungeon_town then self.allow_sell = false end
end

function _M.stockPostFilter(o)
  -- Stock piles of some items.
  local t = _M.stock_levels[o.type] or _M.stock_levels[o.subtype]
  while type(t) == 'string' do t = stock_levels[t] end
  if type(t) == 'table' then
    -- TODO Don't multi-stock ego items.
    local size = 1
    for _, tt in ipairs(t) do
      if o.cost <= tt.cost then size = size + rng.dice(tt.dice, tt.sides) end
    end
    o:setNumber(size)
  end

  return true
end

function _M:getObjectPrice(who, o, buying)
  local price = o:getPrice(false)
  if price < 0 then return 0 end

  local name = o:getName()
  local adj = who:getStatRelated('cha_price')
  if self.owner then
    adj = adj + self.owner.inflate + self.owner:getRaceAdjust(who)
  end
  if buying then
    adj = math.max(100, 100 + (adj - 300))
    -- Black market prices suck.
    if self.black_market then price = price*2 end
  else
    -- T2 describes this as a "Mega Hack^3"
    if o.type == 'ammo' then price = math.floor(price/5) end
    adj = math.min(100, 100 + (300 - adj))
    -- Black market prices suck.
    if self.black_market then price = math.floor(price/2) end
    if self.owner then price = math.min(price, self.owner.max_cost) end
  end

  local p = math.max(1, math.floor((price*adj + 50)/100))
  return math.max(1, math.floor((price*adj + 50)/100))
end

function _M:tryBuy(who, o, item, qty)
  local price = self:getObjectPrice(who, o, true)
  if who.gold >= price*qty then
    return qty, price*qty
  else
    Dialog:simplePopup('Not enough gold', 'Liar!  You have not the gold!')
  end
end

function _M:trySell(who, o, item, qty)
  local price = self:getObjectPrice(who, o, false)
  -- I thought I remembered the stores having a purse that they could run
  -- out of, but I can't find it in the code.
  return qty, price*qty
end

function _M:onBuy(who, o, item, qty, before)
  if before then return end
  local price = self:getObjectPrice(who, o, true)
  if who.gold >= price*qty then
    who:adjGold(-price*qty)
    local o_name = o:getName{ force_qty=qty }
    game.log('%s bought %s for %s gold.', who.name, o_name, price)
  end
end

function _M:doBuy(who, o, item, qty, store_dialog)
  qty = math.min(qty, o:getNumber())
  local price
  qty, price = self:tryBuy(who, o, item, qty)
  if qty then
    local o_name = o:getName{ force_qty=qty }
    local function callback(ok)
      if (ok) then
	self:onBuy(who, o, item, qty, true)
	self:transfer(self, who, item, qty)
	self:onBuy(who, o, item, qty, false)
	if store_dialog then store_dialog:updateStore() end
      end
    end
    Dialog:yesnoPopup('Buy', ('Buy %s for %d gold?'):format(o_name, price), callback, 'Buy', 'Cancel')
  end
end

function _M:doSell(who, o, item, qty, store_dialog)
  qty = math.min(qty, o:getNumber())
  local price
  qty, price = self:trySell(who, o, item, qty)
  if qty then
    local o_name = o:getName{ force_qty=qty }
    local function callback(ok)
      if (ok) then
	self:onSell(who, o, item, qty, true)
	self:transfer(who, self, item, qty)
	self:onSell(who, o, item, nb, false)
	if store_dialog then store_dialog:updateStore() end
      end
    end
    Dialog:yesnoPopup('Sell', ('Sell %s for %d gold?'):format(o_name, price), callback, 'Sell', 'Cancel')
  end
end

function _M:descObjectPrice(who, what, o)
  return ('%d'):format(self:getObjectPrice(who, o, what == 'buy'))
end
