-- $Id: ActorStats.lua 73 2012-09-20 01:22:45Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'

module(..., package.seeall, class.make)

_M.stats = { 'STR', 'INT', 'WIS', 'DEX', 'CON', 'CHA' }
_M.stat_names = {
  STR = 'Strength',
  INT = 'Intelligence',
  WIS = 'Wisdom',
  DEX = 'Dexterity',
  CON = 'Constitution',
  CHA = 'Charisma',
}

-- Base stat value is represented internally as an integer from 0 (=3) to
-- 250 (=18/100).  For non-formatting purposes, we divide by ten on the way
-- out.  This means a birth or equipment adjustment of +1 can always be
-- applied by adding 10 to the internal base value.

function _M:init(t)
  self.stats = {}
  for _, s in ipairs(_M.stats) do
    self.stats[s] = { cur = 70, base = 70, birth_adj = 0, equip_adj = 0 }
  end
end

function _M:getStat(s, t)
  t = t or {}
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)

  if t.raw then
    assert(stat[t.raw], "no '"..t.raw.."' field for stat "..s)
    return stat[t.raw]
  end

  local val = t.base and stat.base or stat.cur
  if not t.no_birth then val = val + stat.birth_adj*10 end
  if not t.no_equip then val = val + stat.equip_adj*10 end
  if t.delta then val = val + t.delta*10 end
  val = 30 + util.bound(val, 0, 370)

  if t.fmt then
    return self.formatStat(val, t.fmt)
  else
    -- For non-display purposes, the last digit of an 18/NNN stat has no
    -- effect.
    val = math.floor(val/10)
    if t.scale then
      val = math.floor(((val - 3)*t.scale)/37)
    elseif t.index then
      val = val - 3
    end
    return val
  end
end

function _M.formatStat(val, fmt)
  if fmt == 'default' then
    fmt = config.settings.tome2.linear_stats and 'linear' or 'pctile'
  end
  if val <= 180 or fmt == 'linear' then
    -- Slight Hack(TM):  18/01 base stat with a birth adjustment of -3 will
    -- be displayed as 15.1 even in pctile display mode.
    local v = math.floor(val/10)
    if val % 10 == 0 then
      return string.format('%2d', v)
    else
      return string.format('%2d.%1d', v, val % 10)
    end
  end
  if val < 280 then return string.format("18/%02d", val - 180) end
  if val < 400 then return string.format("18/%3d", val - 180) end
  return '18/***'
end

function _M:setStat(s, t)
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)
  assert(t and type(t) == 'table', 'missing or malformed adjustment spec for stat '..s)

  for field, val in pairs(t) do
    assert(stat[field], "no '"..field.."' adjustment for stat "..s)
    stat[field] = val
    self.changed = true
  end
end

function _M:isStatMaxed(s)
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)
  return stat.cur >= 250
end

function _M:isStatDrained(s)
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)
  return stat.cur < stat.base
end

function _M:bumpStatUp(s)
  local stat = self.stats[s]
  assert(stat, 'no stat '..s)
  local restored = stat.cur < stat.base
  stat.cur = stat.base
  local val = stat.cur
  if val < 150 then
    -- < 18 gets bumped up 1 point, 25% chance of extra point
    val = val + (rng.range(1, 4) == 1 and 20 or 10)
  elseif val < 248 then
    -- >= 18, < 18/98 gets 1/6 to 1/3 of the diff from max pctile, but not
    -- above 18/99.
    local delta = math.max(1, math.floor(((250 - val)/2 + 3)/2))
    val = math.max(249, val + rng.range(1, delta) + math.floor(delta/2))
  elseif val < 250 then
    -- 18/98 and 18/99 get one pctile point.
    val = val + 1
  end
  local boosted = val > stat.cur
  stat.cur = val
  stat.base = val
  return boosted, restored
end

local related = {
  con_maxhp = {
    stat = 'CON',
    default = 0,
    values = {
      [0] = -5,
      [1] = -3,
      [2] = -2,
      [3] = -1,
      [12] = 1,
      [13] = 1,
      [14] = 2,
      [15] = 3,
      [16] = 4,
      [17] = 4,
      [18] = 4,
      [19] = 4,
      [20] = 5,
      [21] = 6,
      [22] = 7,
      [23] = 8,
      [24] = 9,
      [25] = 10,
      [26] = 11,
      [27] = 12,
      [28] = 13,
      [29] = 14,
      [30] = 15,
      [31] = 16,
      [32] = 18,
      [33] = 20,
      [34] = 22,
      [35] = 25,
      [36] = 26,
      [37] = 27,
    },
  },
  int_wis_mana = {
    stat = function(self)
      local int = self:getStat('INT', {index=true})
      local wis = self:getStat('WIS', {index=true})
      return int > wis and 'INT' or 'WIS'
    end,
    default = 0,
    values = {
      [5] = 1,
      [6] = 2,
      [7] = 2,
      [8] = 2,
      [9] = 2,
      [10] = 2,
      [11] = 2,
      [12] = 2,
      [13] = 2,
      [14] = 2,
      [15] = 3,
      [16] = 3,
      [17] = 3,
      [18] = 3,
      [19] = 3,
      [20] = 4,
      [21] = 4,
      [22] = 5,
      [23] = 6,
      [24] = 7,
      [25] = 8,
      [26] = 9,
      [27] = 10,
      [28] = 11,
      [29] = 12,
      [30] = 13,
      [31] = 14,
      [32] = 15,
      [33] = 16,
      [34] = 16,
      [35] = 17,
      [36] = 17,
      [37] = 18,
    },
  },
  dex_to_hit = {
    stat = 'DEX',
    default = 0,
    values = {
      [0] = -3,
      [1] = -2,
      [2] = -2,
      [3] = -1,
      [4] = -1,
      [13] = 1,
      [14] = 2,
      [15] = 3,
      [16] = 3,
      [17] = 3,
      [18] = 3,
      [19] = 3,
      [20] = 4,
      [21] = 4,
      [22] = 4,
      [23] = 4,
      [24] = 5,
      [25] = 6,
      [26] = 7,
      [27] = 8,
      [28] = 9,
      [29] = 9,
      [30] = 10,
      [31] = 11,
      [32] = 12,
      [33] = 13,
      [34] = 14,
      [35] = 15,
      [36] = 15,
      [37] = 16,
    },
  },
  str_to_hit = {
    stat = 'STR',
    default = 0,
    values = {
      [0] = -3,
      [1] = -2,
      [2] = -1,
      [3] = -1,
      [15] = 1,
      [16] = 1,
      [17] = 1,
      [18] = 1,
      [19] = 1,
      [20] = 1,
      [21] = 1,
      [22] = 2,
      [23] = 3,
      [24] = 4,
      [25] = 5,
      [26] = 6,
      [27] = 7,
      [28] = 8,
      [29] = 9,
      [30] = 10,
      [31] = 11,
      [32] = 12,
      [33] = 13,
      [34] = 14,
      [35] = 15,
      [36] = 15,
      [37] = 16,
    },
  },
  int_disarm = {
    stat = 'INT',
    default = 0,
    values = {
      [5] = 1,
      [6] = 1,
      [7] = 1,
      [8] = 1,
      [9] = 1,
      [10] = 1,
      [11] = 1,
      [12] = 2,
      [13] = 2,
      [14] = 2,
      [15] = 3,
      [16] = 3,
      [17] = 3,
      [18] = 4,
      [19] = 4,
      [20] = 5,
      [21] = 6,
      [22] = 7,
      [23] = 8,
      [24] = 9,
      [25] = 10,
      [26] = 10,
      [27] = 11,
      [28] = 12,
      [29] = 13,
      [30] = 14,
      [31] = 15,
      [32] = 16,
      [33] = 17,
      [34] = 18,
      [35] = 19,
      [36] = 19,
      [37] = 20,
    },
  },
  dex_disarm = {
    stat = 'DEX',
    default = 0,
    values = {
      [10] = 1,
      [11] = 1,
      [12] = 1,
      [13] = 2,
      [14] = 2,
      [15] = 4,
      [16] = 4,
      [17] = 4,
      [18] = 4,
      [19] = 5,
      [20] = 5,
      [21] = 5,
      [22] = 6,
      [23] = 6,
      [24] = 7,
      [25] = 8,
      [26] = 8,
      [27] = 8,
      [28] = 8,
      [29] = 8,
      [30] = 9,
      [31] = 9,
      [32] = 9,
      [33] = 9,
      [34] = 9,
      [35] = 10,
      [36] = 10,
      [37] = 10,
    },
  },
  dex_save = {
    stat = 'DEX',
    values = {
      [0] = 0,
      [1] = 1,
      [2] = 2,
      [3] = 3,
      [4] = 4,
      [5] = 5,
      [6] = 5,
      [7] = 6,
      [8] = 6,
      [9] = 7,
      [10] = 7,
      [11] = 8,
      [12] = 8,
      [13] = 9,
      [14] = 9,
      [15] = 10,
      [16] = 10,
      [17] = 15,
      [18] = 15,
      [19] = 20,
      [20] = 25,
      [21] = 30,
      [22] = 35,
      [23] = 40,
      [24] = 45,
      [25] = 50,
      [26] = 60,
      [27] = 70,
      [28] = 80,
      [29] = 90,
      [30] = 100,
      [31] = 100,
      [32] = 100,
      [33] = 100,
      [34] = 100,
      [35] = 100,
      [36] = 100,
      [37] = 100,
    },
  },
  cha_price = {
    stat = 'CHA',
    values = {
      [0] = 130,
      [1] = 125,
      [2] = 122,
      [3] = 120,
      [4] = 118,
      [5] = 116,
      [6] = 114,
      [7] = 112,
      [8] = 110,
      [9] = 108,
      [10] = 106,
      [11] = 104,
      [12] = 103,
      [13] = 102,
      [14] = 101,
      [15] = 100,
      [16] = 99,
      [17] = 98,
      [18] = 97,
      [19] = 96,
      [20] = 95,
      [21] = 94,
      [22] = 93,
      [23] = 92,
      [24] = 91,
      [25] = 90,
      [26] = 89,
      [27] = 88,
      [28] = 87,
      [29] = 86,
      [30] = 85,
      [31] = 84,
      [32] = 83,
      [33] = 82,
      [34] = 81,
      [35] = 80,
      [36] = 79,
      [37] = 78,
    }
  },
}

function _M:getStatRelated(tag)
  assert(related[tag], 'no such related value '..tag)
  if type(related[tag]) == 'function' then
    return related[tag](self)
  elseif type(related[tag] == 'table') then
    local def = related[tag]
    local stat = def.stat
    if type(def.stat) == 'function' then stat = stat(self) end
    local idx = self:getStat(stat, {index=true})
    return def.values[idx] or def.default or 0
  else
    return 0
  end
end
