-- $Id: PlayerCombat.lua 23 2012-08-18 17:31:59Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local DamageType = require 'engine.DamageType'

module(..., package.seeall, class.make)

function _M:init(t)
  self.melee_style = t.melee_style or 'weapon'
end

function _M.oneBlowWithWeapon(src, target, wpn)
  local combat = wpn.combat or { 1, 1, 0, 0 }
  local dam = rng.dice(combat[1], combat[2]) + combat[4]
  dam = dam + (src.to_d or 0) + (src.to_d_melee or 0)
  local dt = DamageType:get(DamageType.PHYSICAL)
  dt.projector(src, target.x, target.y, DamageType.PHYSICAL, math.max(0, dam))
end

local wpn_blows = {
  { 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3 },
  { 1, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4 },
  { 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5 },
  { 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 5 },
  { 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5 },
  { 2, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 6 },
  { 2, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 6 },
  { 2, 3, 3, 4, 4, 4, 5, 5, 5, 5, 5, 6 },
  { 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 6, 6 },
  { 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6 },
  { 3, 3, 4, 4, 4, 4, 5, 5, 5, 6, 6, 6 },
  { 3, 3, 4, 4, 4, 4, 5, 5, 6, 6, 6, 6 },
}
function _M.numBlowsWithWeapon(src, wpn)
  -- Dummy weapon has no weight, and gets one blow.
  if not wpn.weight then return 1 end

  local str = src:getStat('STR')
  local dex = src:getStat('DEX')
  local div = math.max(wpn.weight, src.weapon_blows.base_wgt)*10
  local mul = src.weapon_blows.wgt_mult
  local s_idx = math.floor(((str <= 17 and str or (str - 16)*10)*mul)/div)
  s_idx = util.bound(s_idx, 0, 11)
  local d_idx = dex <= 9  and 0 or
		dex <= 18 and 1 or
		dex <= 22 and 2 or
		dex <= 26 and math.floor((dex - 23)/2) or
		dex <= 34 and dex - 24 or
		dex <= 38 and (dex - 19)*2 or 20
  d_idx = util.bound(d_idx, 0, 11)
  local max = 4 + (src.abilities.EXTRA_MAX_BLOWS_1 and 1 or 0) +
		  (src.abilities.EXTRA_MAX_BLOWS_2 and 1 or 0)
  return math.min(max, wpn_blows[s_idx+1][d_idx+1])
end

function _M.oneBlowSpecial(src, target, skill, dummy_wpn, blows)
  local sk_lev = src:getSkillLevel(skill)
  local dt = DamageType:get(DamageType.PHYSICAL)

  local ok_blows = {}
  -- Collect the blows we can do.
  for _, blow in ipairs(blows) do
    if blow.min_level <= sk_lev then table.insert(ok_blows, blow) end
  end
  local blow = nil
  if #ok_blows == 0 then
    blow = blows[1]
  else
    -- Try a few times and keep the best blow type we get.
    -- (TODO Don't retry if stunned or confused)
    for _ = 1, math.max(1, math.floor(sk_lev/7)) do
      local try
      -- Higher-level blows are less likely.
      repeat try = rng.table(ok_blows) until rng.range(1, sk_lev) >= try.chance
      if not blow or try.min_level > blow.min_level then blow = try end
    end
  end

  local dam = rng.dice(blow.dam[1], blow.dam[2]) + dummy_wpn.combat[4]
  dam = dam + (src.to_d or 0) + (src.to_d_melee or 0)
  local p = { atk_msg=blow.msg }

  -- TODO apply effects
  -- TODO check for critical hit
  dt.projector(src, target.x, target.y, DamageType.PHYSICAL, math.max(0, dam), p)
end

local barehand_blows = {
  {
    msg = " punches %s",
    min_level = 1,
    chance = 0,
    dam = { 2, 4 },
    effects = {},
  },
  {
    msg = " kicks %s",
    min_level = 2,
    chance = 0,
    dam = { 2, 6 },
    effects = {},
  },
  {
    msg = " strikes %s",
    min_level = 3,
    chance = 0,
    dam = { 2, 7 },
    effects = {},
  },
  {
    msg = " hits %s with a knee",
    min_level = 5,
    chance = 5,
    dam = { 4, 3 },
    effects = { KNEE=0 },
  },
  {
    msg = " hits %s with an elbow",
    min_level = 7,
    chance = 5,
    dam = { 2, 8 },
    effects = {},
  },
  {
    msg = " butts %s",
    min_level = 9,
    chance = 10,
    dam = { 4, 5 },
    effects = {},
  },
  {
    msg = " kicks %s",
    min_level = 11,
    chance = 10,
    dam = { 6, 4 },
    effects = { SLOW=0 },
  },
  {
    msg = " uppercuts %s",
    min_level = 13,
    chance = 12,
    dam = { 8, 4 },
    effects = { STUN=6 },
  },
  {
    msg = " double-kicks %s",
    min_level = 16,
    chance = 15,
    dam = { 10, 4 },
    effects = { STUN=8 },
  },
  {
    msg = " hits %s with a Cat's Claw",
    min_level = 20,
    chance = 20,
    dam = { 10, 5 },
    effects = {},
  },
  {
    msg = " hits %s with a jump kick",
    min_level = 25,
    chance = 25,
    dam = { 10, 6 },
    effects = { STUN=10 },
  },
  {
    msg = " hits %s with an Eagle's Claw",
    min_level = 29,
    chance = 25,
    dam = { 12, 6 },
    effects = {},
  },
  {
    msg = " hits %s with a circle kick",
    min_level = 33,
    chance = 30,
    dam = { 12, 8 },
    effects = { STUN=10 },
  },
  {
    msg = " hits %s with an Iron Fist",
    min_level = 37,
    chance = 35,
    dam = { 16, 8 },
    effects = { STUN=10 },
  },
  {
    msg = " hits %s with a flying kick",
    min_level = 41,
    chance = 35,
    dam = { 16, 10 },
    effects = { STUN=12 },
  },
  {
    msg = " hits %s with a Dragon Fist",
    min_level = 45,
    chance = 35,
    dam = { 20, 10 },
    effects = { STUN=16 },
  },
  {
    msg = " hits %s with a Crushing Blow",
    min_level = 48,
    chance = 35,
    dam = { 20, 12 },
    effects = { STUN=18 },
  },
}
function _M.oneBlowBarehand(src, target, dummy_wpn)
  return _M.oneBlowSpecial(src, target, 'BAREHAND_COMBAT', dummy_wpn, barehand_blows)
end

local bearform_blows = {
  {
    msg = " claws %s",
    min_level = 1,
    chance = 0,
    dam = { 3, 4 },
    effects = { STUN=4 },
  },
  {
    msg = " swats %s",
    min_level = 4,
    chance = 0,
    dam = { 4, 4 },
    effects = { WOUND=20 },
  },
  {
    msg = " bites %s",
    min_level = 9,
    chance = 2,
    dam = { 4, 4 },
    effects = { WOUND=30 },
  },
  {
    msg = " hugs %s",
    min_level = 15,
    chance = 5,
    dam = { 6, 4 },
    effects = { FULL_SLOW=0 },
  },
  {
    msg = " swats and rakes %s",
    min_level = 25,
    chance = 10,
    dam = { 6, 5 },
    effects = { STUN=10, WOUND=10 },
  },
  {
    msg = " hugs and claws %s",
    min_level = 30,
    chance = 15,
    dam = { 6, 6 },
    effects = { FULL_SLOW=60, WOUND=60 },
  },
  {
    msg = " double swats %s",
    min_level = 35,
    chance = 20,
    dam = { 9, 7 },
    effects = { STUN=20, WOUND=20 },
  },
  {
    msg = " double swats and rakes %s",
    min_level = 40,
    chance = 25,
    dam = { 10, 10 },
    effects = { STUN=25, WOUND=25 },
  },
}
function _M.oneBlowBearform(src, target, dummy_wpn)
  return _M.oneBlowSpecial(src, target, 'BEARFORM_COMBAT', dummy_wpn, bearform_blows)
end

local armor_slots = { 'BODY', 'CLOAK', 'OFF_HAND', 'GLOVES', 'BOOTS', 'HELM' }
function _M:armorWeight()
  local total = 0
  for _, slot in ipairs(armor_slots) do
    for _, o in ipairs(self.inven[slot] or {}) do
      total = total + o.weight
    end
  end
  return total
end

-- TODO Would it be to over-elaborate to break melee styles out into /data?
_M.melee_styles = {
  weapon = {
    desc = 'Weapon combat',
    skill = 'WEAPONMASTERY',
    one_blow = _M.oneBlowWithWeapon,
    num_blows = _M.numBlowsWithWeapon,
    all_weapons = function(src)
      local wpns = src:getInven('MAIN_HAND')
      -- Dummy weapon if we're not wielding anything.
      if #wpns == 0 then return { { combat = { 1, 1, 0, 0 } } } end
      return wpns
    end
  },
  barehand = {
    desc = 'Barehanded combat',
    skill = 'BAREHAND_COMBAT',
    one_blow = _M.oneBlowBarehand,
    num_blows = function(src, _)
      local wpns = src:getInven('MAIN_HAND')
      if wpns and #wpns > 0 then return 1 end
      local sk = src:getSkillLevel('BAREHAND_COMBAT')
      -- This gives us 1=1/10=2/20=3/30=4/35=5/40=6/45=7/50=8
      local bonus = 1 + math.floor(sk/10) + math.max(0, math.floor((sk-25)/10))
      if src:armorWeight() > 10 + sk*4/10 then
	bonus = math.floor(bonus/2)
      end
      return 1 + bonus
    end,
    all_weapons = function(src)
      local sk = src:getSkillLevel('BAREHAND_COMBAT')
      local wpns = src:getInven('MAIN_HAND')
      local th, td = 0, 0
      if (not wpns or #wpns == 0) and src:armorWeight() > 10 + sk*4/10 then
	th = math.floor(sk/3)
	td = th
      end
      return { { combat = { 1, 1, th, td } } }
    end
  },
  bearform = {
    desc = 'Bearform combat',
    skill = 'BEARFORM_COMBAT',
    one_blow = _M.oneBlowBearform,
    num_blows = function(src, _)
      return 2 + src:getSkillLevel('BEARFORM_COMBAT', 5)
    end,
    all_weapons = function(src)
      local sk = src:getSkillLevel('BEARFORM_COMBAT')
      return { { combat = { 1, 1, -math.floor(sk/5), math.floor(sk/2) } } }
    end
  },
}
local melee_style_order = { 'weapon', 'barehand', 'bearform' }

function _M:selectMeleeStyle()
  local orig_style = self.melee_style
  local styles = {}
  for _, style in ipairs(melee_style_order) do
    local def = self.melee_styles[style]
    -- TODO Hide BEARFORM_COMBAT skill while not in bear form.
    if self:getSkillLevel(def.skill) > 0 then
      table.insert(styles, { name=def.desc, tag=style })
    end
  end
  local callback = function(sel)
    if not sel or not sel.tag then return end
    -- TODO Unwield any wielded weapons if we switch away from weapon
    -- combat.
    self.melee_style = sel.tag
    self.changed = true
    if self.melee_style ~= orig_style then
      self:useEnergy()
    end
    self:recalcEverything()
  end
  Dialog:listPopup('Melee Style', 'Choose style:', styles, 300, 100, callback)
end

function _M:getSkillForMeleeStyle()
  local melee_style = self.melee_style or 'weapon'
  return melee_styles[melee_style].skill
end

local skill_by_wpn_subtype = {
  sword = 'SWORD_MASTERY',
  axe = 'AXE_MASTERY',
  hafted = 'HAFTED_MASTERY',
  polearm = 'POLEARM_MASTERY',
}
function _M:getSkillForWeapon()
  local wpns = self:getInven('MAIN_HAND')
  if #wpns == 0 then return nil end
  local sk = nil
  for _, wpn in ipairs(wpns) do
    if sk and sk ~= skill_by_wpn_subtype[wpn.subtype] then return nil end
    sk = skill_by_wpn_subtype[wpn.subtype]
  end
  return sk
end

function _M:attackTarget(target)
  local melee_style = self.melee_style or 'weapon'
  local ms = melee_styles[melee_style]
  local wpns = ms.all_weapons(self)

  -- Compute number of blows for each weapon (so in theory, a multi-armed
  -- Possessor wielding Nimthanc and Skullcleaver could get six hits with
  -- the former and four with the latter [which may not be how the original
  -- ToME2 handled it, but...]).
  local n_blows = { total=0 }
  for i, wpn in ipairs(wpns) do
    n_blows[i] = ms.num_blows(self, wpn)
    n_blows.total = n_blows.total + n_blows[i]
  end

  -- Remember this for possible use by Spread Blows.
  local blows = 0

  for i, wpn in ipairs(wpns) do
    if target.dead then break end
    for _ = 1, n_blows[i] do
      blows = blows + 1
      local th = self.to_h + self.to_h_melee + wpn.combat[3]
      local chance = self.skill_fighting + th*3
      if self:checkHit(target, chance) then
	ms.one_blow(self, target, wpn)
	if target.dead then break end
      else
	game.log('%s misses %s.', self.name, target.name)
      end
    end
  end

  -- Player speed is apparently factored in by useEnergy().
  local energy = game.energy_to_act
  if melee_style == 'weapon' and self.abilities.SPREAD_BLOWS then
    -- TODO Maybe Spread-Blows-like abilities for barehand and bearform
    -- combat.
    energy = energy * blows/n_blows.total
  end
  self:useEnergy(energy)
end
