-- $Id: PlayerFates.lua 79 2012-09-20 04:24:03Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'
local Dialog = require 'engine.Dialog'

module(..., package.seeall, class.make)

_G.registerOnTurn(function()
  if config.settings.tome2.fates and game.player.level > 10 and rng.range(1, 50000) == 1
  then
    game.player:gainFate()
  end
end)

-- This roughly duplicates the pre-resolving part of Zone:makeEntity() with
-- prob_filter=true.
local function pick_entity_target(f, type, base_list, filter, name_flags)
  local cb = function(e) return game.zone:checkFilter(e, filter, type) end
  -- Amazingly Dodgy Hack(TM):  Zone:computeRarities() expects an actual
  -- Level object; I think we can fake it with this.
  local fake_level = {
    level = f.level,
    getEntitlesList = function(x) return {} end,
  }
  local list = game.zone:computeRarities(type, base_list, fake_level, cb, filter.add_levels, filter.special_rarity)
  for _ = 1, 50 do
    local e = game.zone:pickEntity(list)
    if e then
      f.target_type = type
      f.target_def = e.define_as
      f.target_desc = e:getName(name_flags)
      return true
    end
  end
  print("[FATE] Couldn't find an "..type.." to generate")
  return false
end

_M.fate_types = {
  find_obj = {
    maybe = true,
    pick_target = function(f)
      local filter = { not_properties = { special_gene, insta_art, norm_art } }
      return pick_entity_target(f, 'object', game.zone.object_list, filter, { force_qty=1 })
    end,
  },
  find_mon = {
    maybe = true,
    pick_target = function(f)
      return pick_entity_target(f, 'actor', game.zone.npc_list, {}, { indef_art=true })
    end,
  },
  find_art = {
    maybe = true,
    -- TODO Actually implement this.
    pick_target = function(f) f.target_desc = 'an artifact'; return true end,
  },
  no_die_mortal = {
    no_level = true,
    desc = function(f)
      return 'You are fated never to die by the hand of a mortal being.'
    end,
  },
  die = {
    desc = function(f)
      return 'You are fated to die on level '..f.level..'.'
    end,
  },
}

function _M:init(t)
  self.fates = {}
end

function _M:gainFate()
  -- If currently lucky, don't get a death fate.
  local i = rng.range(1, self.luck > 0 and 17 or 18)
  local f = {}
  if i <= 7 then
    f.act = 'find_obj'
  elseif i <= 14 then
    f.act = 'find_mon'
  elseif i <= 16 then
    f.act = 'find_art'
  elseif i == 17 then
    f.act = 'no_die_mortal'
  elseif i == 18 then
    f.act = 'die'
  end
  local def = self.fate_types[f.act]
  if def.maybe then f.maybe = rng.percent(50) end
  if not def.no_level then
    local zone_max = game.zone.max_level
    local level = rng.range(zone_max - 20, zone_max + 20)
    f.level = util.bound(level, 1, 98)
  end
  if def.pick_target and not def.pick_target(f) then return end
  table.insert(self.fates, f)
  game.log('#VIOLET#More of your prophecy has been unearthed!#LAST#')
  game.log('#VIOLET#You should see a soothsayer quickly.#LAST#')
end

function _M:descFate(f)
  local def = self.fate_types[f.act]
  if def.desc then return def.desc(f) end

  local maybe = f.maybe and 'may' or 'are fated to'
  return 'You '..maybe..' find '..f.target_desc..' on level '..f.level..'.'
end

function _M:learnUnknownFate()
  local unknown = {}
  for _, f in ipairs(self.fates) do
    if not f.known then table.insert(unknown, f) end
  end
  if #unknown == 0 then return nil end
  local f = unknown[rng.range(1, #unknown)]
  f.known = true
  return self:descFate(f)
end

function _M:showFates()
  local text = 'You do not currently have any fates.'
  local any_unknown = false
  print('[FATE] currently '..#self.fates..' fates')
  if #self.fates > 0 then
    local fates = {}
    for _, f in ipairs(self.fates) do
      print('[FATE]   '..(f.known and 'known' or 'unknown')..' "'..self:descFate(f)..'"')
      if f.known then
	table.insert(fates, '• ' .. self:descFate(f))
      else
	any_unknown = true
      end
    end
    if #fates > 0 then text = table.concat(fates, '\n') end
    if #fates == 0 and any_unknown then text = nil end
  end
  if any_unknown then
    text = (text and (text.."\n\n") or '') .. 'Some of your fate is still unknown.  You should visit a soothsayer in town to learn about it.'
  end

  -- Vaguely Annoying Hack(TM):  Dialog.simpleLongPopup() appears to have a
  -- bug with line spacing, so we recreate its guts here with a bugfix.
  local title = 'Fates:  '..self.name
  local w = game.w * 0.5
  local fct, no_leave = nil, nil
  local font = core.display.newFont("/data/font/DroidSans.ttf", 14)
  local list = text:splitLines(w - 10, font)

  local fh = font:lineSkip()
  local th = font:size(title)
  local d = Dialog.new(title, w + 8, th + fh * #list + 50, nil, nil, nil, font)
  if no_leave then
    d:keyCommands{}
  else
    d:keyCommands{__DEFAULT=function() game:unregisterDialog(d) if fct then fct() end end}
    d:mouseZones{{x=0, y=0, w=game.w, h=game.h, norestrict=true, fct=function(b) if b ~= "none" then game:unregisterDialog(d) if fct then fct() end end end, norestrict=true}}
  end
  d.drawDialog = function(self, s)
    local h = 4
    local r, g, b
    for i = 1, #list do
      r, g, b = s:drawColorStringBlended(self.font, list[i], 5, h, r, g, b) h = h + fh
    end
    self.changed = false
  end
  game:registerDialog(d)
end
