-- $Id: GameOptions.lua 84 2012-09-20 23:47:01Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local Tab = require 'engine.ui.Tab'
--local ListColumns = require 'engine.ui.ListColumns'
local TweakedTreeList = require 'mod.dialogs.ui.TweakedTreeList'
local Textzone = require 'engine.ui.Textzone'
local Button = require 'engine.ui.Button'
local GetQuantity = require 'engine.dialogs.GetQuantity'

module(..., package.seeall, class.inherit(Dialog))

local default_desc = '{No description for this option.}'
local placeholder_desc = '\n\n#RED#This option is not yet implemented.  We Apologize for the Inconvenience.™'
local birth_desc = '#YELLOW##{italic}#This is a birth option; its value cannot be changed after the character is created.#{normal}##LAST#\n\n'

_M.opt_yes = string.toTString('#00FF00#Y#LAST#')
_M.opt_no = string.toTString('#FF0000#N#LAST#')
_M.opt_yn = { [true] = _M.opt_yes, [false] = _M.opt_no }

_M.options_def = {}
_M.options_order = {}
_M.opt_grp_order = {}

function _M:loadDefinition(file, env)
  env = env or {
    GameOptions = self,
    newOption = function(t) self:newOption(t) end
  }
  local f, err = util.loadfilemods(file, setmetatable(env, {__index = _G}))
  if not f and err then error(err) end
  f()
end

function _M:newOption(t)
  assert(t.key, 'no option key')
  assert(t.name, 'no option name')
  assert(t.group, 'no option group')
  assert(type(t.default) ~= 'nil', 'no option default')
  if type(t.default) == 'number' then
    assert(t.min and t.max, 'no option min/max value')
  end
  t.desc = t.desc or default_desc
  if t.placeholder then t.desc = t.desc..placeholder_desc end
  self.options_def[t.key] = t
  table.insert(self.options_order, t.key)
  if not self.options_order[t.group] then
    table.insert(self.opt_grp_order, t.group)
    self.options_order[t.group] = {}
  end
  table.insert(self.options_order[t.group], t.key)
end

function _M:setupOptions()
  config.settings.tome2 = config.settings.tome2 or {}
  for key, info in pairs(self.options_def) do
    -- This both fills out an empty settings table and fills in default
    -- values for any new options added since the last time settings were
    -- saved.
    if type(config.settings.tome2[key]) == 'nil' then
      config.settings.tome2[key] = info.default
    end
  end
end

function _M:saveOptions()
  local settings = 'tome2 = {\n';
  for key, val in pairs(config.settings.tome2) do
    if type(val) == 'boolean' then
      settings = settings .. ('  %s = %s,\n'):format(key, val and 'true' or 'false')
    elseif type(val) == 'number' then
      settings = settings .. ('  %s = %d,\n'):format(key, val)
    elseif type(val) == 'string' then
      settings = settings .. ('  %s = %q,\n'):format(key, val)
    else
      print('[OPTIONS] unrecognized option type', key, val)
    end
  end
  game:saveSettings('tome2', settings .. '}\n')
end

function _M:init()
  Dialog.init(self, 'Game Options', 600, 400)
  local orig_opt = table.clone(config.settings.tome2)

  local tab_h = 0
  local first_grp
  self.c_tabs = {}
  for _, grp in ipairs(self.opt_grp_order) do
    self.c_tabs[grp] = Tab.new {
      title = string.capitalize(grp),
      default = first_grp and false or true,
      fct = function() end,
      on_change = function(s) if s then self:switchTo(grp) end end,
    }
    tab_h = math.max(tab_h, self.c_tabs[grp].h)
    if not first_grp then first_grp = grp end
  end

  self.c_ok = Button.new {
    text = 'Apply',
    fct = function()
      game:unregisterDialog(self)
      self:saveOptions()
      game.player.changed = true
    end
  }
  self.c_cancel = Button.new {
    text = 'Cancel',
    fct = function()
      config.settings.tome2 = orig_opt
      game:unregisterDialog(self)
      game.player.changed = true
    end
  }

  local main_h = self.ih - tab_h - self.c_ok.h
  self.c_desc = Textzone.new {
    width = math.floor(self.iw/2 - 10),
    height = main_h,
    scrollbar = true,
    no_color_bleed = true,
    text = ''
  }
  self.c_list = self:createOptionsList(math.floor(self.iw/2 - 10), main_h, self.c_desc)
  self.c_list:populateForGroup(first_grp)

  local ui = {}
  local prev_tab = 0
  for _, grp in ipairs(self.opt_grp_order) do
    table.insert(ui, { left=prev_tab, top=0, ui=self.c_tabs[grp] })
    prev_tab = self.c_tabs[grp]
  end
  table.insert(ui, { left=0, top=tab_h, ui=self.c_list })
  table.insert(ui, { right=0, top=tab_h, ui=self.c_desc })
  table.insert(ui, { right=0, bottom=0, ui=self.c_cancel })
  table.insert(ui, { right=self.c_cancel, bottom=0, ui=self.c_ok })
  self:loadUI(ui)
  self:setFocus(self.c_list)
  self:setupUI()
  self.key:addBinds { EXIT = self.c_cancel.fct }
end

function _M:switchTo(grp)
  for key, tab in pairs(self.c_tabs) do tab.selected = key == grp end
  self.c_list:populateForGroup(grp)
  local item = self.c_list.list[self.c_list.sel]
  if not item then
    self.c_list.sel = 1
    item = self.c_list.list[1]
  end
  self.c_desc.text = self:computeOptionDesc(item.key, birther)
  self.c_desc:generate()
end

-- [static] Create a TreeList object that manipulates game options and
-- puts text descriptions in the provided Textzone.
function _M:createOptionsList(w, h, text, set_cb, birther)
  local c_list
  --c_list = ListColumns.new {
  c_list = TweakedTreeList.new {
    simple_list = true,
    width = w,
    height = h,
    columns = {
      { name='Option', width=90, display_prop='name' },
      { name=' ', width=10, display_prop='display_val' },
    },
    --list = {},
    tree = {},
    fct = function(item, _, _)
      self:alterOption(item, c_list, set_cb)
    end,
    select = function(item, sel)
      text.text = self:computeOptionDesc(item.key, birther)
      text:generate()
    end
  }
  if birther then
    local opts = {}
    for _, key in ipairs(self.options_order) do
      local def = self.options_def[key]
      if def.group == 'birth' or def.show_in_birther then
	table.insert(opts, key)
      end
    end
    self:generateList(opts, c_list)
  else
    c_list.populateForGroup = function(list_self, grp)
      self:generateList(self.options_order[grp], list_self)
    end
  end
  return c_list
end

-- [static] Called from the TreeList created by createOptionsList() to
-- populate the list with the provided options.
function _M:generateList(keys, c_list)
  local list = {}

  for _, key in ipairs(keys) do
    local def = self.options_def[key]
    local v = config.settings.tome2[key]
    local dv = v
    if type(v) == 'boolean' then
      dv = self.opt_yn[v]
    elseif def.short_values then
      dv = def.short_values[v]
    end

    list[#list+1] = { key=key, name=def.name, value=v, display_val=dv }
  end

  --c_list.list = list
  c_list.tree = list
  c_list:generate()
end

-- [static] Called from the TreeList created by createOptionsList() to
-- actually set option values.
function _M:alterOption(item, c_list, set_cb)
  if not item then return end
  local def = self.options_def[item.key]
  if type(item.value) == 'boolean' then
    item.value = not item.value
    item.display_val = self.opt_yn[item.value]
    config.settings.tome2[item.key] = item.value
    c_list:drawItem(item)
    if set_cb then set_cb(item.key) end
  elseif type(item.value) == 'number' then
    local min = self.options_def[item.key].min
    local max = self.options_def[item.key].max
    local callback = function(qty)
      qty = util.bound(qty, min, max)
      item.value = qty
      item.display_val = qty
      config.settings.tome2[item.key] = qty
      c_list:drawItem(item)
      if set_cb then set_cb(item.key) end
    end
    local d = GetQuantity.new(item.name, item.name, item.value, max, callback, min)
    game:registerDialog(d)
  elseif def.values then
    local vals = {}
    for _, val in ipairs(def.values) do
      table.insert(vals, { name = val })
    end
    local callback = function(sel)
      if not sel or not sel.name then return end
      item.value = sel.name
      item.display_val = def.short_values and def.short_values[item.value] or item.value
      config.settings.tome2[item.key] = item.value
      c_list:drawItem(item)
      if set_cb then set_cb(item.key) end
    end
    Dialog:listPopup(def.name, 'Select value:', vals, 300, 100, callback)
  end
end

-- [static] Called from the TreeList created by createOptionsList() to
-- create the description of the provided option.
function _M:computeOptionDesc(key, birther)
  local def = self.options_def[key]
  if def.group == 'birth' and not birther then
    return birth_desc .. def.desc
  else
    return def.desc
  end
end
