-- $Id: SkillDialog.lua 17 2012-07-12 02:52:35Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright (C) 2012 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local ListColumns = require 'engine.ui.ListColumns'
local TweakedTreeList = require 'mod.dialogs.ui.TweakedTreeList'
local Separator = require 'engine.ui.Separator'
local Textzone = require 'engine.ui.Textzone'
local PlayerSkills = require 'mod.class.interface.PlayerSkills'
local ActorStats = require 'mod.class.interface.ActorStats'

module(..., package.seeall, class.inherit(Dialog))

local points_text = 'Unused skill points: #%s#%d#LAST#'
local function getPointsText(pts)
  local c = pts > 0 and '00FF00' or 'SLATE'
  return points_text:format(c, pts)
end

local instructions = [[Keyboard: #00FF00#up/down arrow#FFFFFF# to select a skill or ability; #00FF00#right arrow#FFFFFF# to increase a skill or choose an ability; #00FF00#left arrow#FFFFFF# to decrease a skill or unchoose an ability.
Mouse: #00FF00#left click#FFFFFF# to increase a skill or choose an ability; #00FF00#right click#LAST# to decrease a skill or unchoose an ability.
#00FF00#RETURN#FFFFFF# to accept, #00FF00#ESC#FFFFFF# to cancel.]]

function _M:init(actor, on_finish)
  self.actor = actor
  self.actor_dup = actor:clone()
  self.on_finish = on_finish
  self.temp_skill = {}
  self.temp_abil = {}

  local w = math.min(game.w*0.8, 800)
  local h = math.floor(game.h*0.8)
  local x = math.floor((game.w - w)/2)
  local y = math.floor((game.h - h)/2)
  Dialog.init(self, 'Skill points: '..actor.name, w, h, x, y)

  local left_w = math.floor((2*self.iw)/3 - 10)
  local right_w = math.floor(self.iw/3 - 10)

  self.c_tut = Textzone.new {
    width = right_w,
    auto_height = true,
    no_color_bleed = true,
    text = instructions,
  }

  self.hs = Separator.new { dir='horizontal', size=self.ih }
  self.vs_l = Separator.new { dir='vertical', size=left_w }
  self.vs_r = Separator.new { dir='vertical', size=right_w }

  self.c_points = Textzone.new {
    width = left_w,
    auto_height = true,
    no_color_bleed = true,
    text = getPointsText(self.actor_dup.skill_points),
  }
  self.c_info = Textzone.new {
    width = right_w,
    height = self.ih - self.c_tut.h - self.vs_r.h,
    scrollbar = true,
    no_color_bleed = true,
    text = '',
  }

  local left_avail_h = self.ih - self.c_points.h - self.vs_l.h
  local sk_tree_h = math.floor(left_avail_h*2/3 - 10)
  local ab_list_h = math.floor(left_avail_h/3 - 10)

  self:generateSkills()
  self.c_sk_tree = TweakedTreeList.new {
    width = left_w,
    height = sk_tree_h,
    all_clicks = true,
    scrollbar = true,
    columns = {
      { name='Skill', width=70, display_prop='disp_name' },
      { name='Level', width=15, display_prop='disp_value' },
      { name='Mult',  width=15, display_prop='disp_mult' },
    },
    tree = self.skill_tree,
    fct = function(item, sel, v) self:adjSkill(v == 'left' and 1 or -1) end,
    select = function(item, sel)
      self.sel_skill = item
      self:onSelect(item)
    end,
    on_expand = function(item) end,
    on_drawitem = function(item) end,
  }
  self.c_sk_tree.key:addBinds {
    MOVE_LEFT = function() self:adjSkill(-1) end,
    MOVE_RIGHT = function() self:adjSkill(1) end,
  }

  self:generateAbilities()
  self.c_ab_list = ListColumns.new {
    width = left_w,
    height = ab_list_h,
    all_click = true,
    scrollbar = true,
    columns = {
      { name='Ability', width=85, display_prop='disp_name' },
      { name='Cost',    width=15, display_prop='disp_cost' },
    },
    list = self.abil_list,
    fct = function(item, sel, v) self:adjAbil(v == 'left') end,
    select = function(item, sel)
      self.sel_abil = item
      self:onSelect(item)
    end,
  }
  self.c_ab_list.key:addBinds {
    MOVE_LEFT = function() self:adjAbil(-1) end,
    MOVE_RIGHT = function() self:adjAbil(1) end,
  }

  -- We want the main dialog to handle accept
  self.c_sk_tree.key:removeBind('ACCEPT')
  self.c_ab_list.key:removeBind('ACCEPT')

  self:loadUI {
    { left=0, top=0, ui=self.c_points },
    { left=0, top=self.c_points, ui=self.vs_l },
    { left=0, top=self.vs_l, ui=self.c_sk_tree },
    { left=0, bottom=0, ui=self.c_ab_list },
    { left=self.c_sk_tree, top=0, ui=self.hs },
    { right=0, top=0, ui=self.c_tut },
    { right=0, top=self.c_tut.h, ui=self.vs_r },
    { right=0, bottom=0, ui=self.c_info },
  }
  self:setFocus(self.c_sk_tree)
  self:setupUI()

  self.key:addBind('ACCEPT', function() self:applyDialog() end)
  self.key:addBind('EXIT', function() self:exitDialog() end)
end

function _M:isShowSkill(tag)
  local sk = self.actor_dup.skills[tag]
  if sk.mod > 0 then return true end
  for _, child in ipairs(sk.children or {}) do
    if self:isShowSkill(child) then return true end
  end
  return false
end

function _M:generateSkillSubtree(tag)
  local def = PlayerSkills.skills_def[tag]
  local sk = self.actor_dup.skills[tag]
  local tmp_pts = self.temp_skill[tag] or 0
  local fmt = '%6.3f'

  local name = def.name
  local value = fmt:format(math.min(sk.value/PlayerSkills.skill_step, PlayerSkills.skill_max))
  local mod = fmt:format(sk.mod/PlayerSkills.skill_step)
  if tmp_pts > 0 then name = '#00FF00#'..name..'#WHITE#' end
  if sk.value >= PlayerSkills.skill_max then
    value = '#00FFFF#'..value..'#WHITE#'
    mod = '#SLATE#'..mod..'#WHITE#'
    if tmp_pts == 0 then name = '#00FFFF#'..name..'#WHITE#' end
  elseif tmp_pts > 0 then
    value = '#00FF00#'..value..'#WHITE#'
  end

  local item = {
    tag = tag,
    skill = sk,
    disp_name = name:toTString(),
    desc = sk.desc,
    value = sk.value,
    disp_value = value:toTString(),
    mod = sk.mod,
    disp_mult = mod:toTString(),
  }
  if def.children then
    local nodes = {}
    for _, child in ipairs(def.children) do
      if self:isShowSkill(child) then
	nodes[#nodes+1] = self:generateSkillSubtree(child)
      end
    end
    if #nodes > 0 then
      item.nodes = nodes
      item.shown = true
    end
  end
  if self.sel_skill and self.sel_skill.tag == tag then
    self.sel_skill = item
  end
  return item
end

function _M:generateSkills()
  local tree = {}
  for _, tag in ipairs(PlayerSkills.skills_toplevel) do
    if self:isShowSkill(tag) then
      tree[#tree+1] = self:generateSkillSubtree(tag)
    end
  end

  self.skill_tree = tree
  if self.c_sk_tree then
    self.c_sk_tree.tree = tree
    self.c_sk_tree:generate()
    -- Grmbl mutter... generate() just clobbered our key bindings from init().
    self.c_sk_tree.key:removeBind('ACCEPT')
    self.c_sk_tree.key:addBinds {
      MOVE_LEFT = function() self:adjSkill(-1) end,
      MOVE_RIGHT = function() self:adjSkill(1) end,
    }
    if self.sel_skill and self.c_sk_tree.focused then
      self:onSelect(self.sel_skill)
    end
  end
end

function _M:generateAbilities()
  local list = {}
  local newsel = nil
  for _, tag in ipairs(PlayerSkills.abilities_order) do
    local def = PlayerSkills.abilities_def[tag]
    local name = def.name
    local cost = ''..def.cost
    if self.actor_dup.abilities[tag] then
      name = '#00FFFF#'..name..'#WHITE#'
      cost = '#00FFFF#'..cost..'#WHITE#'
    elseif self.temp_abil[tag] then
      name = '#00FF00#'..name..'#WHITE#'
      cost = '#00FF00#'..cost..'#WHITE#'
    elseif not self:isPrereqMet(tag) then
      cost = '#SLATE#'..cost..'#WHITE#'
    end
    local item = {
      tag = tag,
      cost = def.cost,
      desc = def.desc,
      disp_name = name:toTString(),
      disp_cost = cost:toTString(),
      pre = true,
    }
    list[#list+1] = item
    if self.sel_abil and self.sel_abil.tag == tag then newsel = item end
  end

  self.abil_list = list
  if newsel then self.sel_abil = newsel end
  if self.c_ab_list then
    self.c_ab_list.list = list
    self.c_ab_list:generate();
    -- Grmbl mutter... generate() just clobbered our key bindings from init().
    self.c_ab_list.key:removeBind('ACCEPT')
    self.c_ab_list.key:addBinds {
      MOVE_LEFT = function() self:adjAbil(-1) end,
      MOVE_RIGHT = function() self:adjAbil(1) end,
    }
    if self.sel_abil and self.c_ab_list.focused then
      self:onSelect(self.sel_abil)
    end
  end
end

function _M:isPrereqMet(tag)
  local def = PlayerSkills.abilities_def[tag]
  if not def.prereq then return true end
  for sk, val in pairs(def.prereq.skill or {}) do
    if self.actor_dup:getSkillLevel(sk) < val then return false end
  end
  for st, val in pairs(def.prereq.stat or {}) do
    if self.actor_dup:getStat(st) < val then return false end
  end
  for _, ab in ipairs(def.prereq.ability or {}) do
    if not self.actor_dup.abilities[ab] and not self.temp_abil[ab] then
      return false
    end
  end
  return true
end

function _M:adjSkill(delta)
  if not self.sel_skill then return end
  local sk = self.sel_skill
  local tag = self.sel_skill.tag
  local pts = self.temp_skill[tag] or 0

  if delta < 0 then
    if pts + delta < 0 then
      self:simplePopup('Cannot decrease', 'You have not assigned any points to this skill.')
      return
    end
  else
    if self.actor_dup.skill_points - delta < 0 then
      self:simplePopup('Cannot increase', 'You have no unused skill points to spend.')
      return
    elseif sk.value >= PlayerSkills.skill_max then
      self:simplePopup('Cannot increase', 'This skill is already at its maximum value.')
      return
    elseif sk.value + delta*sk.mod >= (self.actor_dup.level + 5)*PlayerSkills.skill_step then
      self:simplePopup('Cannot increase', 'Cannot increase skill above level + 4.')
      return
    elseif sk.mod <= 0 then
      self:simplePopup('Cannot increase', 'You have not learned this skill.')
      return
    end
  end

  local sk_a = self.actor_dup.skills[tag]
  local def = PlayerSkills.skills_def[tag]
  sk_a.value = sk.value + delta*sk.mod
  for rtag, amt in pairs(def.ripple_to or {}) do
    local sk_ar = self.actor_dup.skills[rtag]
    sk_ar.value = sk_ar.value + delta*math.floor(sk_ar.mod*amt/100)
  end

  self.temp_skill[tag] = pts + delta
  self.actor_dup.skill_points = self.actor_dup.skill_points - delta
  self:generateSkills()
  self.c_points.text = getPointsText(self.actor_dup.skill_points);
  self.c_points:generate()
end

function _M:adjAbil(choose)
  if not self.sel_abil then return end
  local tag = self.sel_abil.tag
  local cost = self.sel_abil.cost
  if choose and self.actor_dup[tag] then
    self:simplePopup('Known ability', 'You already know this ability.')
    return
  elseif choose and self.temp_abil[tag] then
    self:simplePopup('Selected ability', 'You have already chosen to learn this ability.')
    return
  elseif choose and cost > self.actor_dup.skill_points then
    self:simplePopup('Skill points', 'You do not have enough skill points to learn this ability.')
    return
  elseif choose and not self:isPrereqMet(tag) then
    self:simplePopup('Prerequisites', 'You have not met the prerequisites to learn this ability.')
    return
  elseif not choose and not self.temp_abil[tag] and self.actor_dup[tag] then
    self:simplePopup('Known ability', 'You cannot unlearn this ability.')
    return
  elseif not choose and not self.temp_abil[tag] and not self.actor_dup[tag] then
    self:simplePopup('Known ability', 'You have not chosen to learn this ability.')
    return
  end

  self.temp_abil[tag] = choose
  self.actor_dup.skill_points = self.actor_dup.skill_points + (choose and -cost or cost)
  self:generateAbilities()
  self.c_points.text = getPointsText(self.actor_dup.skill_points);
  self.c_points:generate()
end

function _M:onSelect(item)
  local text = item.desc
  if item.pre then
    local def = PlayerSkills.abilities_def[item.tag]
    if def.prereq then
      text = text .. '\n\nPrerequisites:'
      for st, val in pairs(def.prereq.stat or {}) do
	local c = self.actor_dup:getStat(st) >= val and '00FF00' or 'FF0000'
	text = text .. ("\n#%s#• %s %s#LAST#"):format(c, st, ActorStats.formatStat(val*10, 'default'))
      end
      for sk, val in pairs(def.prereq.skill or {}) do
	local sk_def = PlayerSkills.skills_def[sk]
	local c = self.actor_dup:getSkillLevel(sk) >= val and '00FF00' or 'FF0000'
	text = text .. ("\n#%s#• %s %d#LAST#"):format(c, sk_def.name, val)
      end
      for _, ab in ipairs(def.prereq.ability or {}) do
	local ab_def = PlayerSkills.abilities_def[ab]
	local c = (self.actor_dup.abilities[ab] or self.temp_abil[ab]) and '00FF00' or 'FF0000'
	text = text .. ("\n#%s#• %s#LAST#"):format(c, ab_def.name)
      end
    end
  end

  self.c_info.text = text
  self.c_info:generate()
end

function _M:exitDialog()
  game:unregisterDialog(self)
  self.actor_dup = {}
  if self.on_finish then self.on_finish() end
end

function _M:applyDialog()
  if self.actor.skill_points == self.actor_dup.skill_points then
    -- No skill points spent == no changes to apply.
    self:exitDialog()
    return
  end

  local callback = function(yes, cancel)
    if cancel then return end
    if yes then
      for tag, pts in pairs(self.temp_skill) do
	if pts > 0 then
	  local sk_a, sk_d = self.actor.skills[tag], self.actor_dup.skills[tag]
	  sk_a.value = math.min(sk_d.value, PlayerSkills.skill_max)
	end
      end
      for tag, _ in pairs(self.temp_abil) do
	self.actor.abilities[tag] = true
	self.actor.bought_abilities[tag] = true
      end
      self.actor.skill_points = self.actor_dup.skill_points
    end
    self:exitDialog()
  end
  self:yesnocancelPopup('Finish', 'Apply the requested changes?', callback)
end
